# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
Library of useful functions for interpolation
"""
import numpy as np
from numpy.polynomial.legendre import leggauss
import numba
from numba.typed import List
import astropy.units as u

@numba.njit(nogil=True,fastmath=True)
def bilinear_interpolation(z00, z10, z01, z11, x, y):
    """
    2D interpolation
    Applies linear interpolation across x and y between xmin,xmax and ymin,ymax

    Parameters
    ----------
        z00: array
            Array corresponding to xmin,ymin
        z10: array
            Array corresponding to xmax,ymin
        z01: array
            Array corresponding to xmin,ymax
        z11: array
            Array corresponding to xmax,ymax
        x: float
            weight on x coord
        y: float
            weight on y coord
    """
    xy = x*y
    res = np.zeros_like(z00)
    for i in range(z00.shape[0]):
        res[i] = (z11[i]-z01[i]+z00[i]-z10[i])*xy +(z01[i]-z00[i])*y +(z10[i]-z00[i])*x +z00[i]
    return res

@numba.njit(nogil=True,fastmath=True)
def bilinear_interpolation_array(z00, z10, z01, z11, x, y):
    """
    2D interpolation
    Applies linear interpolation across x and y between xmin,xmax and ymin,ymax.
    For this variant of bilinear_interpolation (mainly for chemistry),
    x, y and z must be 1D-arrays of the same length.

    Parameters
    ----------
        z00: array
            Array corresponding to xmin,ymin
        z10: array
            Array corresponding to xmax,ymin
        z01: array
            Array corresponding to xmin,ymax
        z11: array
            Array corresponding to xmax,ymax
        x: array
            weights on x coord
        y: array
            weights on y coord
    """
    xy = x*y
    res = np.zeros_like(z00)
    for i in range(z00.shape[0]):
        res[i] = (z11[i]-z01[i]+z00[i]-z10[i])*xy[i] +(z01[i]-z00[i])*y[i] +(z10[i]-z00[i])*x[i] +z00[i]
    return res


@numba.njit(nogil=True,fastmath=True)
def linear_interpolation(z00, z10, x):
    """1D interpolation.

    Applies linear interpolation in x between xmin, xmax.

    Parameters
    ----------
        z00: array
            Array corresponding to xmin
        z10: array
            Array corresponding to xmax
        x: float
            weight on x coord
    """
    res = np.zeros_like(z00)
    for i in range(z00.shape[0]):
#        res[i] = z10[i]*x+z00[i]*x2
        tmp1=z00[i]
        tmp2=z10[i]
        tmp2-=tmp1
        tmp1+=tmp2*x
        res[i] = tmp1
    return res


def interp_ind_weights(x_to_interp,x_grid):
    """Finds the indices and weights to interpolate from a x_grid to a x_to_interp.
    """
    xmin=x_grid.min()
    xmax=x_grid.max()
    used_x=np.where(x_to_interp>xmax,xmax,x_to_interp)
    used_x=np.where(used_x<xmin,xmin,used_x)
    indices=np.searchsorted(x_grid,used_x)
    indices=np.where(indices==0,1,indices)
    return indices,(used_x-x_grid[indices-1])/(x_grid[indices]-x_grid[indices-1])

def rebin_ind_weights(old_bin_grid, new_bin_grid):
    """Computes the indices and weights to be used when
    rebinning an array of values f of size N in the bins delimited by old_bin_grid (of size N+1)
    onto new bins delimited by new_bin_grid (of size Nnew+1)
    Returns indices, weights to be used as follows to proceed with the rebinning

    >>> f_rebinned=[np.dot(f[indicestosum[ii]-1:indicestosum[ii+1]],final_weights[ii])
    for ii in range(Nnew)]
    """
#    new_bin_grid_used=np.where(new_bin_grid<old_bin_grid[0],old_bin_grid[0],new_bin_grid)
    indicestosum=np.searchsorted(old_bin_grid, new_bin_grid, side='right')
    if indicestosum[-1]==old_bin_grid.size : indicestosum[-1]-=1
    dg=old_bin_grid[1:]-old_bin_grid[:-1]
    final_weights=List()
    for ig in range(new_bin_grid.size-1):
        if indicestosum[ig+1]>indicestosum[ig]:
            weights=np.concatenate((old_bin_grid[[indicestosum[ig]]]-new_bin_grid[[ig]],\
                dg[indicestosum[ig]:indicestosum[ig+1]-1],  \
                new_bin_grid[[ig+1]]-old_bin_grid[[indicestosum[ig+1]-1]]))
        else:
            weights=np.array([new_bin_grid[ig+1]-new_bin_grid[ig]])
        weights=weights/np.sum(weights)
        final_weights.append(weights)
    return indicestosum, final_weights

@numba.njit()
def kdata_conv_loop(kdata1,kdata2,kdataconv,shape):
    """Computes the convolution of two kdata tables.

    Parameters
    ----------
        kdata1,kdata2 : arrays
            The two ktable.kdata tables to convolve.
        shape : array
            shape of the ktabke.kdata tables to convolve (last size is Ng).
        kdataconv : array
            Result table where the last dimension as a length equal to Ng^2.
    """
    Ng=shape[-1]
    for i in range(shape[0]):
     for j in range(shape[1]):
      for k in range(shape[2]):
       for l in range(Ng):
        for m in range(Ng):
            kdataconv[i,j,k,l*Ng+m]=kdata1[i,j,k,m]+kdata2[i,j,k,l]

@numba.njit(nogil=True,fastmath=True)
def kdata_conv_loop_profile(kdata1,kdata2,kdataconv,Nlay,Nw,Ng):
    """Computes the convolution of two atmospheric kdata profiles.

    Nothing is returned. kdataconv is changed in place. 

    Parameters
    ----------
        kdata1,kdata2 : arrays
            The two ktable.kdata tables to convolve.
        kdataconv : array
            Result table where the last dimension as a length equal to Ng^2.
        Nlay : int
            Number of atmospheric layers
        Nw : int
            Number of wavenumber points
        Ng : int
            Number of g-points
    """
    for i in range(Nlay):
        for j in range(Nw):
            for l in range(Ng):
                for m in range(Ng):
                    kdataconv[i,j,l*Ng+m]=kdata1[i,j,m]+kdata2[i,j,l]

@numba.njit(fastmath=True)
def rebin(f_fine,fine_grid,coarse_grid):
    """Computes the binned version of a function on a coarser grid.
    The two grids do not need to have the same boundaries.

    Parameters
    ----------
        f_fine: array
            Function to be rebinned, given on the fine_grid.
        fine_grid: array
            The high resolution grid edges we start with.
        coarse_grid: array
            The coarser resolution grid edges inside which we want to bin the function f.
       """
    indicestosum=np.searchsorted(fine_grid,coarse_grid,side='right')
    Nfine=fine_grid.size
    if indicestosum[-1]==Nfine : 
        indicestosum=np.where(indicestosum<Nfine,indicestosum,Nfine-1)
    dgrid=np.diff(fine_grid)
    Nn=coarse_grid.size-1
    f_coarse=np.zeros(Nn)
    ifine=indicestosum[0]-1
    for ii in range(Nn):
      if indicestosum[ii+1]>indicestosum[ii]:
        tmp_w=fine_grid[ifine+1]-coarse_grid[ii]
        tmp_f=f_fine[ifine]*tmp_w
        for iifine in range(indicestosum[ii],indicestosum[ii+1]-1):
          tmp_f+=dgrid[iifine]*f_fine[iifine]
          tmp_w+=dgrid[iifine]
        ifine=indicestosum[ii+1]-1
        tmp_dw=(coarse_grid[ii+1]-fine_grid[ifine])
        #print(ifine,tmp_dw,f_fine[ifine])
        tmp_f+=f_fine[ifine]*tmp_dw
        tmp_w+=tmp_dw
        f_coarse[ii]=tmp_f/tmp_w
      else:
        f_coarse[ii]=f_fine[ifine]
    return f_coarse

@numba.njit(fastmath=True)
def g_sample_4d(new_ggrid, new_kdata, old_ggrid, old_kdata):
    """Reinterpolte the g grid inplace.
    """
    shape=old_kdata.shape
    for iP in range(shape[0]):
      for iT in range(shape[1]):
        for iW in range(shape[2]):
            new_kdata[iP,iT,iW,:]=np.interp(new_ggrid, old_ggrid, old_kdata[iP,iT,iW,:])
    return

@numba.njit(fastmath=True)
def g_sample_5d(new_ggrid, new_kdata, old_ggrid, old_kdata):
    """Reinterpolte the g grid inplace.
    """
    shape=old_kdata.shape
    for iP in range(shape[0]):
      for iT in range(shape[1]):
        for iX in range(shape[2]):
          for iW in range(shape[3]):
            new_kdata[iP,iT,iX,iW,:]=np.interp(new_ggrid, old_ggrid, old_kdata[iP,iT,iX,iW,:])
    return

@numba.njit(nogil=True,fastmath=True)
def RandOverlap_2_kdata_prof(Nlay, Nw, Ng, kdata1, kdata2, weights, ggrid):
    """Function to randomely mix the opacities of 2 species in an atmospheric profile.

    Parameters
    ----------
        Nlay, Nw, Ng: int
            Number of layers, spectral bins, and gauss points. 
        kdata1, kdata2: arrays of size (Nlay, Nw, Ng)
            vmr weighted cross-sections for the two species.
        weights: array
            gauss weights.
        ggrid: array
            g-points.
    
    Returns
    -------
        array
            k-coefficient array of the mix over the atmospheric column.
    """
    kdataconv=np.zeros((Nlay,Nw,Ng**2))
    weightsconv=np.zeros(Ng**2)
    newkdata=np.zeros((Nlay,Nw,Ng))

    for jj in range(Ng):
        for ii in range(Ng):
            weightsconv[ii*Ng+jj]=weights[jj]*weights[ii]

    kdata_conv_loop_profile(kdata1,kdata2,kdataconv,Nlay,Nw,Ng)

    for ilay in range(Nlay):
        for iW in range(Nw):
            if (kdata1[ilay,iW,-1]-kdata2[ilay,iW,0])*(kdata2[ilay,iW,-1]-kdata1[ilay,iW,0])<0.:
                #ii+=1
                if kdata1[ilay,iW,-1] > kdata2[ilay,iW,-1]:
                    newkdata[ilay,iW,:] = kdata1[ilay,iW,:] 
                else:
                    newkdata[ilay,iW,:] = kdata2[ilay,iW,:]
                continue
            tmp=kdataconv[ilay,iW,:]
            indices=np.argsort(tmp)
            kdatasort=tmp[indices]
            weightssort=weightsconv[indices]
            newggrid=np.cumsum(weightssort)
            #ind=np.searchsorted(newggrid,ggrid,side='left')
            #newkdata[ilay,iW,:]=kdatasort[ind]
            newkdata[ilay,iW,:]=np.interp(ggrid,newggrid,kdatasort)
    return newkdata

def unit_convert(quantity,unit_file='unspecified',unit_in='unspecified',unit_out='unspecified'):
    """Chooses the final unit to use and the conversion factor to apply.

    Parameters
    ----------
        quantity: str
            The name of the pysical quantity handled for potential error messages
        unit_file/unit_in/unit_out: str

            Respectively:

              * String with the unit found in (or assumed from the format of) the initial data,
              * The unit we think the initial data are in if unit_file is 'unspecified'
                or (we believe) wrong,
              * The unit we want to convert to. If unspecified, we do not convert.
    Returns
    -------
        unit_to_write: str
            Resulting unit.
        conversion_factor: float
            A multiplicating factor for the data to proceed to the conversion.
    """
    if unit_in!='unspecified':
      if((unit_file!='unspecified') and (unit_file!=unit_in)):
        print('Be careful, you are assuming that '+quantity+' is '+unit_in)
        print('but the input file says that it is '+unit_file+'. The former will be used')
      starting_unit=unit_in
    else:
      if unit_file=='unspecified':
        raise NotImplementedError("""I could not find the {quantity} used in the file.
            So you should tell me what unit you think is used by specifying 
            the keyword argument: file_{quantity}= a unit recognized by the astropy.units library.
            If you want to convert to another unit, add the {quantity} keyword
            with the desired unit.""".format(quantity=quantity))
      else:
        starting_unit=unit_file
    if unit_out=='unspecified':
        unit_to_write=starting_unit
    else:
        unit_to_write=unit_out
    return unit_to_write,u.Unit(starting_unit).to(u.Unit(unit_to_write))

def rm_molec(unit_name):
    """Removes "/molecule" or "/molec" from a unit string.

    Parameters
    ----------
        unit_name: str
            String to be changed.
        
    Returns
    -------
        str
            The unit name without the ending "/molecule" or "/molec"
    """
    return unit_name.replace('/molecule','').replace('/molec','')

@numba.njit
def is_sorted(a):
    """Finds out if an array is sorted. Returns True if it is.
    """
    for i in range(a.size-1):
         if a[i+1] < a[i] :
               return False
    return True

def gauss_legendre(order):
    """Computes the weights and abscissa for a Gauss Legendre quadrature of order `order`

    Parameters
    ----------
        order: int
            Order of the quadrature wanted.

    Returns
    -------
        weights: array(order)
            Weights to be used in the quadrature.
        ggrid: array(order)
            Abscissa to be used for the quadrature.
        gedges: array(order+1)
            Cumulative sum of the weights. Goes from 0 to 1.
       """ 
    ggrid,weights=leggauss(order)
    weights=weights/2.
    ggrid=(ggrid+1.)/2.
    gedges=np.insert(np.cumsum(weights),0,0.)
    return weights,ggrid,gedges

def split_gauss_legendre(order, g_split):
    """Computes the weights and abscissa for a split Gauss Legendre quadrature of order `order`:
    Half the points will be put between 0 and `g_split`, and half between `g_split`and 1.

    (This si what is used in petitRADTRANS with g_split=0.9)

    Parameters
    ----------
        order: int
            Order of the quadrature wanted. Needs to be even.
        g_split: float between 0 and 1
            Splitting point.

    Returns
    -------
        weights: array(order)
            Weights to be used in the quadrature.
        ggrid: array(order)
            Abscissa to be used for the quadrature.
        gedges: array(order+1)
            Cumulative sum of the weights. Goes from 0 to 1.
       """ 
    if order%2==1:
        print('order should be an even number')
        raise RuntimeError()
    ggrid,weights=leggauss(order//2)
    weights1=weights*g_split/2.
    ggrid1=(ggrid+1.)/2.*g_split
    weights2=weights*(1-g_split)/2.
    ggrid2=(ggrid+1.)/2.*(1-g_split)+g_split
    weights=np.concatenate([weights1,weights2])
    ggrid=np.concatenate([ggrid1,ggrid2])
    gedges=np.insert(np.cumsum(weights),0,0.)
    return weights,ggrid,gedges

@numba.njit
def spectrum_to_kdist(k_hr,wn_hr,dwn_hr,wnedges,ggrid):
    """Creates the k-distribution from a single high resolution spectrum
    
    Parameters
    ----------
        k_hr : array
            Spectrum
        wn_hr : array
            Wavenumber grid
        dwn_hr : array
            Width of the high resolution wavenumber bins.
        wnedges : array
            Lower resolution wavenumber bin edges inside which the k-dist will be computed.
        ggrid : array
            Grid of g-point abscissas

    Returns
    -------
        kdata : array
            k coefficients for each (Wn, g) bin.
    """
    pos=np.searchsorted(wn_hr,wnedges)
    #print(pos)
    kdata=np.zeros((wnedges.size-1,ggrid.size))
    for ib in range(wnedges.size-1):
        if pos[ib+1]==pos[ib]:
            continue
            # JL20 why did I write that ??
            # for the moment, this leaves zeros but we should probably make the code stop
        tmp_k=k_hr[pos[ib]:pos[ib+1]]
        totwg=np.sum(dwn_hr[pos[ib]:pos[ib+1]])
        wg=dwn_hr[pos[ib]:pos[ib+1]]/totwg
        sort_ind=np.argsort(tmp_k)
        newgfine=np.cumsum(wg[sort_ind])
        # choosing the line below or the two after actually makes a significant difference
        kdata[ib]=np.interp(ggrid,newgfine,tmp_k[sort_ind])
        #g_pos=np.searchsorted(newgfine,ggrid,side='right')
        #kdata[ib]=tmp_k[sort_ind[g_pos]] 
        # an interpolation between tmp_k[sort_ind[g_pos]] and tmp_k[sort_ind[g_pos-1]] 
        # would probably be cleaner.
    return kdata



@numba.njit()
def bin_down_corrk_numba(newshape, kdata, old_ggrid, new_ggrid, gedges, indicestosum, \
        wngrid_filter, wn_weigths, num, use_rebin):
    """bins down a kcoefficient table (see :func:`~exo_k.ktable.Ktable.bin_down` for details)
    
    Parameters
    ----------
        newshape : array
            Shape of kdata.
        kdata : array
            table to bin down.
        old_ggrid : array
            Grid of old g-point abscissas for kdata.
        new_ggrid : array
            New g-points for the binned-down k-coefficients.
        gedges : array
            Cumulative sum of the weights. Goes from 0 to 1.
            Used only if use_rebin=True
        indicestosum: list of lists
            Indices of wavenumber bins to be used for the averaging
        wngrid_filter: array
            Indices of the new table where there will actually be data (zero elsewhere)
        wn_weigths: list of lists
            Weights to be used for the averaging (same size as indicestosum)
        num: int
            Number of points to fine sample the g function in log-k space
        use_rebin: boolean
            Whether to use rebin or interp method. 
    """
    ggrid0to1=np.copy(old_ggrid)
    ggrid0to1[0]=0.
    ggrid0to1[-1]=1.
    newkdata=np.zeros(newshape, dtype=np.float64)
    for iW,newiW in enumerate(wngrid_filter[0:-1]):
        tmp_dwn=wn_weigths[iW]
        for iP in range(newshape[0]):
            for iT in range(newshape[1]):            
                tmp_logk=np.log(kdata[iP,iT,indicestosum[iW]-1:indicestosum[iW+1]])
                #if iP==0 and iT==0: print(tmp_logk)
                logk_min=np.amin(tmp_logk[:,0])
                logk_max=np.amax(tmp_logk[:,-1])
                if logk_min==logk_max:
                    newkdata[iP,iT,newiW,:]=np.ones(newshape[-1])*np.exp(logk_max)
                else:
                    logk_max=logk_max+(logk_max-logk_min)/(num-3.)
                    logk_min=logk_min-(logk_max-logk_min)/(num-3.)
                    logkgrid=np.linspace(logk_min,logk_max,num)
                    newg=np.zeros(logkgrid.size)
                    for ii in range(tmp_logk.shape[0]):
                        newg+=np.interp(logkgrid,tmp_logk[ii],ggrid0to1)*tmp_dwn[ii]
                        #newg+=np.interp(logkgrid,tmp_logk[ii],ggrid0to1,
                        #    left=0., right=1.)*tmp_dwn[ii]
                    if use_rebin:
                        newkdata[iP,iT,newiW,:]=rebin(np.exp(logkgrid), newg, gedges)
                    else:
                        newkdata[iP,iT,newiW,:]=np.interp(new_ggrid, newg, np.exp(logkgrid))
    return newkdata

## OLD FUNCTIONS

@numba.njit()
def kdata_conv(kdata1,kdata2,kdataconv,Ng):
    """Deprecated.
    
    Performs the convolution of the kdata over g-space.
    """
    for jj in range(Ng):
        for ii in range(Ng):
            kdataconv[:,:,:,ii*Ng+jj]=kdata1[:,:,:,jj]+kdata2[:,:,:,ii]

@numba.njit()
def kdata_conv_loop_bad(kdata1,kdata2,kdataconv,shape):
    """Deprecated. 
    
    Performs the convolution of the kdata over g-space.
    But the loop is in bad order. Keep for test. 
    """
    Ng=shape[-1]
    for jj in range(Ng):
     for ii in range(Ng):
      for k in range(shape[2]):
       for l in range(shape[1]):
        for m in range(shape[0]):
            kdataconv[m,l,k,ii*Ng+jj]=kdata1[m,l,k,jj]+kdata2[m,l,k,ii]


#@numba.njit(fastmath=True)
def g_interp(logk,Nw,Ng,Num,ggrid,wl_weights,indices):
    """Interpolates logk on a new g-grid. 
    """
    #write = False
    #if write : print(logk);print(Nw,Ng,Num);print(ggrid);print(wl_weights);print(indices)
    lkmin=logk[0,0]
    for iW in range(1,Nw):
        if logk[iW,0] < lkmin : lkmin = logk[iW,0]
    lkmax=logk[0,-1]
    for iW in range(1,Nw):
        if logk[iW,-1] > lkmax : lkmax = logk[iW,-1]
    dlk=(lkmax-lkmin)/(Num-3)
    logkgrid=np.zeros(Num)
    Ovgedges=1./np.diff(ggrid)
    newg=np.zeros(Num)
    lk=lkmin-dlk
    logkgrid[0]=lk
    for iN in range(1,Num):
        lk+=dlk
        logkgrid[iN]=lk
        tmpg=0.
        for iW in range(Nw):
            ind=indices[iW]
            #if write : print(ind,iW,lk,logk[iW,ind])
            if lk < logk[iW,ind]:
                #if write : print('lk<logk')
                if ind!=0 :
                    #if write: print('ind!=0')
                    #if write: print( \
                    # (ggrid[ind]*(lk-logk[iW,ind-1])+ggrid[ind-1]*(logk[iW,ind]-lk))\
                    #   *Ovgedges[ind-1]*wl_weights[iW])
                    tmpg+=(ggrid[ind]*(lk-logk[iW,ind-1])+ggrid[ind-1]*(logk[iW,ind]-lk))  \
                        *Ovgedges[ind-1]*wl_weights[iW]
            else:
                #if write : print('lk>logk')
                if ind!=Ng-1:
                    #if write : print('ind!=Ng-1')
                    ind+=1
                    indices[iW]=ind
                    #if write : print(  \
                    # (ggrid[ind]*(lk-logk[iW,ind-1])+ggrid[ind-1]*(logk[iW,ind]-lk)) \
                    # *Ovgedges[ind-1]*wl_weights[iW])
                    tmpg+=(ggrid[ind]*(lk-logk[iW,ind-1])+ggrid[ind-1]*(logk[iW,ind]-lk)) \
                        *Ovgedges[ind-1]*wl_weights[iW]
                else:
                    #if write : print('ind=Ng-1')
                    tmpg+=wl_weights[iW]
        newg[iN]=tmpg
    #if write : print(newg);print(logkgrid)

    return newg,logkgrid




# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
__init__ module to load the exo_k library
"""
from exo_k.util.user_func import *
from exo_k.two_stream import two_stream_crisp, two_stream_toon, two_stream_lmdz
from exo_k.util.radiation import *
from exo_k.util.interp import gauss_legendre, split_gauss_legendre
from exo_k.util.cst import *
from exo_k.util.spectrum import Spectrum
from exo_k.util.filenames import *
from exo_k.util.molar_mass import Molar_mass
from .ktable import Ktable
from .ktable5d import Ktable5d
from .kdatabase import Kdatabase
from .xtable import Xtable
from .atable import Atable, combine_tables
from .adatabase import Adatabase
from .cia_table import Cia_table
from .ciadatabase import CIAdatabase
from .hires_spectrum import Hires_spectrum
from .atm import Atm_profile, Atm
from .atm_2band import Atm_2band
#from .atm2 import Atm_profile2, Atm2
from .chemistry import EquChemTable, InterpolationChemistry
from .gas_mix import Gas_mix, Known_composite_species
from .settings import Settings
from .rayleigh import Rayleigh



# -*- coding: utf-8 -*-
"""
Created in Jan 2021

@author: jeremy leconte
"""
import numpy as np
import numba

def solve_2stream_nu_xsec(source_nu, dtau_nu, omega0_nu, g_asym_nu,
        flux_top_dw_nu, mu0=0.5, alb_surf=0., flux_at_level=False):
    """Deals with the spectral axis
    """
      
    Nlay,Nw   = dtau_nu.shape
    FLUXUPI_nu   = np.zeros((Nlay+1, Nw))
    FLUXDWI_nu   = np.zeros((Nlay+1, Nw))
    FNETI_nu     = np.zeros((Nlay+1, Nw))
              
# WE NOW ENTER A MAJOR LOOP OVER SPECTRAL INTERVALS IN THE INFRARED
# TO CALCULATE THE NET FLUX IN EACH SPECTRAL INTERVAL
            
    for NW in range(Nw):         
            
# SET UP THE UPPER AND LOWER BOUNDARY CONDITIONS ON THE IR
# CALCULATE THE DOWNWELLING RADIATION AT THE TOP OF THE MODEL
# OR THE TOP LAYER WILL COOL TO SPACE UNPHYSICALLY
            
#            TAUTOP = dtau_nu[1,NW,iG]*PLEV[2]/(PLEV[4]-PLEV[2])
#JL21 simple boundary condition for now to compare with toon
#            TAUTOP = TAUCUMI[0,NW,iG]
#            BTOP   = (1.e0-np.exp(-TAUTOP/mu0))*PLTOP
            
# WE CAN NOW SOLVE FOR THE COEFFICIENTS OF THE TWO STREAM
# CALL A SUBROUTINE THAT SOLVES  FOR THE FLUX TERMS
# WITHIN EACH INTERVAL AT THE MIDPOINT WAVENUMBER 
            
        FMUPI, FMDI, FNETI = solve_2stream(source_nu[:,NW], dtau_nu[:,NW],
                omega0_nu[:,NW], g_asym_nu[:,NW], mu0=mu0,
                flux_top_dw=flux_top_dw_nu[NW], alb_surf=alb_surf,
                flux_at_level=flux_at_level)
        FLUXUPI_nu[:,NW] = FMUPI
        FLUXDWI_nu[:,NW] = FMDI
        FNETI_nu[:,NW] = FNETI
               
    return FLUXUPI_nu, FLUXDWI_nu, FNETI_nu

@numba.jit(nopython=True, fastmath=True, cache=True)
def solve_2stream_nu_corrk(source_nu, dtau_nu, omega0_nu, g_asym_nu,
        flux_top_dw_nu, mu0=0.5, alb_surf=0., flux_at_level=False):
    """Deals with the spectral axis
    """
      
    Nlay,Nw,Ng   = dtau_nu.shape
    FLUXUPI_nu   = np.zeros((Nlay+1, Nw, Ng))
    FLUXDWI_nu   = np.zeros((Nlay+1, Nw, Ng))
    FNETI_nu     = np.zeros((Nlay+1, Nw, Ng))
              
# WE NOW ENTER A MAJOR LOOP OVER SPECTRAL INTERVALS IN THE INFRARED
# TO CALCULATE THE NET FLUX IN EACH SPECTRAL INTERVAL
            
    for NW in range(Nw):         
        for iG in range(Ng):            
            
# SET UP THE UPPER AND LOWER BOUNDARY CONDITIONS ON THE IR
# CALCULATE THE DOWNWELLING RADIATION AT THE TOP OF THE MODEL
# OR THE TOP LAYER WILL COOL TO SPACE UNPHYSICALLY
            
#            TAUTOP = dtau_nu[1,NW,iG]*PLEV[2]/(PLEV[4]-PLEV[2])
#JL21 simple boundary condition for now to compare with toon
#            TAUTOP = TAUCUMI[0,NW,iG]
#            BTOP   = (1.e0-np.exp(-TAUTOP/mu0))*PLTOP
            
# WE CAN NOW SOLVE FOR THE COEFFICIENTS OF THE TWO STREAM
# CALL A SUBROUTINE THAT SOLVES  FOR THE FLUX TERMS
# WITHIN EACH INTERVAL AT THE MIDPOINT WAVENUMBER 
            
            FMUPI, FMDI, FNETI = solve_2stream(source_nu[:,NW], dtau_nu[:,NW,iG],
                    omega0_nu[:,NW,iG], g_asym_nu[:,NW,iG], mu0=mu0,
                    flux_top_dw=flux_top_dw_nu[NW], alb_surf=alb_surf,
                    flux_at_level=flux_at_level)
            FLUXUPI_nu[:,NW,iG] = FMUPI
            FLUXDWI_nu[:,NW,iG] = FMDI
            FNETI_nu[:,NW,iG] = FNETI
               
    return FLUXUPI_nu, FLUXDWI_nu, FNETI_nu

@numba.jit(nopython=True, fastmath=True, cache=True)
def solve_2stream(source, dtau, omega0, g_asym, mu0=0.5, flux_top_dw=0., 
                      alb_surf=0., flux_at_level=False):
    """Based on gfluxi in LMDZ code      
##-----------------------------------------------------------------------
##  THIS SUBROUTINE TAKES THE OPTICAL CONSTANTS AND BOUNDARY CONDITIONS
##  FOR THE INFRARED FLUX AT ONE WAVELENGTH AND SOLVES FOR THE FLUXES AT
##  THE LEVELS.  THIS VERSION IS SET UP TO WORK WITH LAYER OPTICAL DEPTHS
##  MEASURED FROM THE TOP OF EACH LAYER.  THE TOP OF EACH LAYER HAS  
##  OPTICAL DEPTH ZERO.  IN THIS SUB LEVEL N IS ABOVE LAYER N. THAT IS LAYER N
##  HAS LEVEL N ON TOP AND LEVEL N+1 ON BOTTOM. OPTICAL DEPTH INCREASES
##  FROM TOP TO BOTTOM.  SEE C.P. MCKAY, TGM NOTES.
##  THE TRI-DIAGONAL MATRIX SOLVER IS DSOLVER AND IS DOUBLE PRECISION SO MANY 
##  VARIABLES ARE PASSED AS SINGLE THEN BECOME DOUBLE IN DSOLVER
##
## NLL            = NUMBER OF LEVELS (NLAYERS + 1) MUST BE LESS THAT NL (101)
## TLEV(L_LEVELS) = ARRAY OF TEMPERATURES AT GCM LEVELS
## WAVEN          = WAVELENGTH FOR THE COMPUTATION
## DW             = WAVENUMBER INTERVAL
## dtau(NLAYER)   = ARRAY OPTICAL DEPTH OF THE LAYERS
## omega0(NLEVEL)     = SINGLE SCATTERING ALBEDO
## g_asym(NLEVEL) = ASYMMETRY FACTORS, 0=ISOTROPIC
## mu0          = AVERAGE ANGLE, MUST BE EQUAL TO 0.5 IN IR
## alb_surf            = SURFACE REFLECTANCE
## BTOP           = UPPER BOUNDARY CONDITION ON IR INTENSITY (NOT FLUX)
## BSURF          = SURFACE EMISSION = (1-RSFI)*PLANCK, INTENSITY (NOT FLUX)
## FP(NLEVEL)     = UPWARD FLUX AT LEVELS
## FM(NLEVEL)     = DOWNWARD FLUX AT LEVELS
## FMIDP(NLAYER)  = UPWARD FLUX AT LAYER MIDPOINTS
## FMIDM(NLAYER)  = DOWNWARD FLUX AT LAYER MIDPOINTS
##-----------------------------------------------------------------------
"""

##=======================================================================
##     WE GO WITH THE HEMISPHERIC CONSTANT APPROACH IN THE INFRARED
#      
    omeg_max=0.9999999999999
    TAUMAX=8.
    Nlay=dtau.size
    FMIDP=np.zeros((Nlay+1))
    FMIDM=np.zeros((Nlay+1))
    ALPHA=np.zeros((Nlay))
    LAMDA=np.zeros((Nlay))
    omega_tmp=np.zeros((Nlay))
    B0=np.zeros((Nlay))
    B1=np.zeros((Nlay))
    GAMA=np.zeros((Nlay))
    CPM1=np.zeros((Nlay))
    CMM1=np.zeros((Nlay))
    CP=np.zeros((Nlay))
    CM=np.zeros((Nlay))
    E1=np.zeros((Nlay))
    E2=np.zeros((Nlay))
    E3=np.zeros((Nlay))
    E4=np.zeros((Nlay))

    for L in range(Nlay):

        if omega0[L]>=omeg_max:
           omega_tmp[L] = omeg_max
        else:
           omega_tmp[L] = omega0[L]
        
        ALPHA[L] = np.sqrt((1.e0-omega_tmp[L])/(1.e0-omega_tmp[L]*g_asym[L]) )
        LAMDA[L] = ALPHA[L]*(1.e0-omega_tmp[L]*g_asym[L])/mu0
        
        B0[L] = source[L]
        B1[L] = (source[L+1] - B0[L]) / dtau[L]

        GAMA[L] = (1.e0-ALPHA[L])/(1.e0+ALPHA[L])
        TERM    = mu0/(1.e0-omega_tmp[L]*g_asym[L])
         
# CPM1 AND CMM1 ARE THE CPLUS AND CMINUS TERMS EVALUATED
# AT THE TOP OF THE LAYER, THAT IS ZERO OPTICAL DEPTH
         
        CPM1[L] = B0[L]+B1[L]*TERM
        CMM1[L] = B0[L]-B1[L]*TERM
         
# CP AND CM ARE THE CPLUS AND CMINUS TERMS EVALUATED AT THE
# BOTTOM OF THE LAYER.  THAT IS AT dtau OPTICAL DEPTH.
# JL18 put CP and CM after the calculation of CPM1 and CMM1 to avoid unecessary calculations. 
         
        CP[L] = CPM1[L] +B1[L]*dtau[L] 
        CM[L] = CMM1[L] +B1[L]*dtau[L] 
      
# NOW CALCULATE THE EXPONENTIAL TERMS NEEDED
# FOR THE TRIDIAGONAL ROTATED LAYERED METHOD
# WARNING IF dtau[J) IS GREATER THAN ABOUT 35 (VAX)
# WE CLIP IT TO AVOID OVERFLOW.
      
    for L in range(Nlay):
        EP    = np.exp( min((LAMDA[L]*dtau[L]),TAUMAX)) # CLIPPED EXPONENTIAL
        EM    = 1.e0/EP
        E1[L] = EP+GAMA[L]*EM
        E2[L] = EP-GAMA[L]*EM
        E3[L] = GAMA[L]*EP+EM
        E4[L] = GAMA[L]*EP-EM
      
# DOUBLE PRECISION TRIDIAGONAL SOLVER
    BTOP=flux_top_dw
    BSURF=(1.-alb_surf)*source[-1]
 
    XK1,XK2 = DSOLVER(Nlay,GAMA,CP,CM,CPM1,CMM1,E1,E2,E3,E4,BTOP,
                BSURF,alb_surf)
      
# NOW WE CALCULATE THE FLUXES
#UPPER LAYER
    #DTAUK = 0.
    TERM  = mu0/(1.e0-omega_tmp[0]*g_asym[0])
    CPMID    = B0[0] +B1[0]*TERM
    CMMID    = B0[0] -B1[0]*TERM
    FMIDP[0] = XK1[0] + GAMA[L]*XK2[0] + CPMID
    FMIDM[0] = XK1[0]*GAMA[L] + XK2[0] + CMMID
      
    if flux_at_level:
        for L in range(Nlay):
            DTAUK = dtau[L]/2. # here we assume levels are at tau=tau_layer/2.
            EP    = np.exp(min(LAMDA[L]*DTAUK,TAUMAX)) # CLIPPED EXPONENTIAL 
            EM    = 1.e0/EP
            TERM  = mu0/(1.e0-omega_tmp[L]*g_asym[L])
         
# CP AND CM ARE THE CPLUS AND CMINUS TERMS EVALUATED AT THE
# TOP OF THE LAYER.  THAT IS AT 0  OPTICAL DEPTH
         
            CPMID    = B0[L]+B1[L]*DTAUK +B1[L]*TERM
            CMMID    = B0[L]+B1[L]*DTAUK -B1[L]*TERM
            FMIDP[L+1] = XK1[L]*EP + GAMA[L]*XK2[L]*EM + CPMID
            FMIDM[L+1] = XK1[L]*EP*GAMA[L] + XK2[L]*EM + CMMID

    else:
        for L in range(1,Nlay):
            DTAUK = 0.
            EP    = np.exp(min(LAMDA[L]*DTAUK,TAUMAX)) # CLIPPED EXPONENTIAL 
            EM    = 1.e0/EP
            TERM  = mu0/(1.e0-omega_tmp[L]*g_asym[L])
         
# CP AND CM ARE THE CPLUS AND CMINUS TERMS EVALUATED AT THE
# TOP OF THE LAYER.  THAT IS AT 0  OPTICAL DEPTH
         
            CPMID    = B0[L]+B1[L]*DTAUK +B1[L]*TERM
            CMMID    = B0[L]+B1[L]*DTAUK -B1[L]*TERM
            FMIDP[L] = XK1[L]*EP + GAMA[L]*XK2[L]*EM + CPMID
            FMIDM[L] = XK1[L]*EP*GAMA[L] + XK2[L]*EM + CMMID


# And now, for the special bottom layer

        L    = Nlay-1
        DTAUK=dtau[L]
        #DTAUK= 0.
        EP   = np.exp(min((LAMDA[L]*DTAUK),TAUMAX)) # CLIPPED EXPONENTIAL 
        EM   = 1.e0/EP
        TERM = mu0/(1.e0-omega_tmp[L]*g_asym[L])

# CP AND CM ARE THE CPLUS AND CMINUS TERMS EVALUATED AT THE
# BOTTOM OF THE LAYER.  THAT IS AT dtau  OPTICAL DEPTH

        CPMID    = B0[L]+B1[L]*DTAUK +B1[L]*TERM
        CMMID    = B0[L]+B1[L]*DTAUK -B1[L]*TERM
        FMIDP[L+1] = XK1[L]*EP + GAMA[L]*XK2[L]*EM + CPMID
        FMIDM[L+1] = XK1[L]*EP*GAMA[L] + XK2[L]*EM + CMMID
 
    return   FMIDP, FMIDM, FMIDP-FMIDM

@numba.jit(nopython=True, fastmath=True, cache=True)
def DSOLVER(NL,GAMA,CP,CM,CPM1,CMM1,E1,E2,E3,E4,BTOP,
                       BSURF,alb_surf):
    """
#  GCM2.0  Feb 2003
#
# DOUBLE PRECISION VERSION OF SOLVER

#*********************************************************
#* THIS SUBROUTINE SOLVES FOR THE COEFFICIENTS OF THE    *
#* TWO STREAM SOLUTION FOR GENERAL BOUNDARY CONDITIONS   *
#* NO ASSUMPTION OF THE DEPENDENCE ON OPTICAL DEPTH OF   *
#* C-PLUS OR C-MINUS HAS BEEN MADE.                      *
#* NL     = NUMBER OF LAYERS IN THE MODEL                *
#* CP     = C-PLUS EVALUATED AT TAO=0 (TOP)              *
#* CM     = C-MINUS EVALUATED AT TAO=0 (TOP)             *
#* CPM1   = C-PLUS  EVALUATED AT TAOSTAR (BOTTOM)        *
#* CMM1   = C-MINUS EVALUATED AT TAOSTAR (BOTTOM)        *
#* EP     = EXP(LAMDA*dtau)                              *
#* EM     = 1/EP                                         *
#* E1     = EP + GAMA *EM                                *
#* E2     = EP - GAMA *EM                                *
#* E3     = GAMA*EP + EM                                 *
#* E4     = GAMA*EP - EM                                 *
#* BTOP   = THE DIFFUSE RADIATION INTO THE MODEL AT TOP  *
#* BSURF  = THE DIFFUSE RADIATION INTO THE MODEL AT      *
#*          THE BOTTOM: INCLUDES EMMISION AND REFLECTION *
#*          OF THE UNATTENUATED PORTION OF THE DIRECT    *
#*          BEAM. BSTAR+alb_surf*FO*EXP(-TAOSTAR/U0)          *
#* alb_surf    = REFLECTIVITY OF THE SURFACE                  *
#* XK1    = COEFFICIENT OF THE POSITIVE EXP TERM         *
#* XK2    = COEFFICIENT OF THE NEGATIVE EXP TERM         *
#*********************************************************
"""
###      PARAMETER (NMAX=201)
#      IMPLICIT REAL*8  (A-H,O-Z)
#      DIMENSION GAMA(NL),CP(NL),CM(NL),CPM1(NL),CMM1(NL),XK1(NL),
#     *          XK2(NL),E1(NL),E2(NL),E3(NL),E4(NL)
#      DIMENSION AF(2*NL),BF(2*NL),CF(2*NL),DF(2*NL),XK(2*NL)
#======================================================================#

    L=2*NL

#    ************MIXED COEFFICENTS**********
#    THIS VERSION AVOIDS SINGULARITIES ASSOC.
#    WITH omega0=0 BY SOLVING FOR XK1+XK2, AND XK1-XK2.
    AF=np.empty((L))
    BF=np.empty((L))
    CF=np.empty((L))
    DF=np.empty((L))
    XK=np.empty((L))
    XK1=np.empty((NL))    
    XK2=np.empty((NL))    

    AF[0] = 0.0
    BF[0] = GAMA[0]+1.
    CF[0] = GAMA[0]-1.
    DF[0] = BTOP-CMM1[0]
    N     = 0
    LM2   = L-2

#     EVEN TERMS
 
    for I in range(1,LM2,2):
    #  DO I=2,LM2,2
        #print(I,N)
        AF[I] = (E1[N]+E3[N])*(GAMA[N+1]-1.)       
        BF[I] = (E2[N]+E4[N])*(GAMA[N+1]-1.)
        CF[I] = 2.0*(1.-GAMA[N+1]**2)
        DF[I] = (GAMA[N+1]-1.) * (CPM1[N+1] - CP[N]) + \
               (1.-GAMA[N+1])* (CM[N]-CMM1[N+1])
        N     = N+1
 
    N   = 0
    LM1 = L-1
    for I in range(2,LM1,2):
        AF[I] = 2.0*(1.-GAMA[N]**2)
        BF[I] = (E1[N]-E3[N])*(1.+GAMA[N+1])
        CF[I] = (E1[N]+E3[N])*(GAMA[N+1]-1.)
        DF[I] = E3[N]*(CPM1[N+1] - CP[N]) + E1[N]*(CM[N] - CMM1[N+1])
        N     = N+1
 
    AF[-1] = E1[-1]-alb_surf*E3[-1]
    BF[-1] = E2[-1]-alb_surf*E4[-1]
    CF[-1] = 0.0
    DF[-1] = BSURF-CP[-1]+alb_surf*CM[-1]

    XK = DTRIDGL(L,AF,BF,CF,DF)
 
#     ***UNMIX THE COEFFICIENTS****

    for N in range(NL):
        XK1[N] = XK[2*N]+XK[2*N+1]
        XK2[N] = XK[2*N]-XK[2*N+1]

#       NOW TEST TO SEE IF XK2 IS REALLY ZERO TO THE LIMIT OF THE
#       MACHINE ACCURACY  = 1 .E -30
#       XK2 IS THE COEFFICEINT OF THE GROWING EXPONENTIAL AND MUST
#       BE TREATED CAREFULLY

        if XK2[N] == 0.0: continue
#        IF (ABS (XK2(N)/XK(2*N-1)) .LT. 1.E-30) XK2(N)=0.0

        if np.abs(XK2[N]/(XK[2*N-1]+1.e-20)) <= 1.E-30:
            XK2[N]=0.0   # For debug only (with -Ktrap=fp option)

    return XK1, XK2

@numba.jit(nopython=True, fastmath=True, cache=True)
def DTRIDGL(L,AF,BF,CF,DF):
    """
    !  GCM2.0  Feb 2003

    !     DOUBLE PRECISION VERSION OF TRIDGL

          DIMENSION AF(L),BF(L),CF(L),DF(L),XK(L)
          DIMENSION AS(2*L),DS(2*L)

    !*    THIS SUBROUTINE SOLVES A SYSTEM OF TRIDIAGIONAL MATRIX
    !*    EQUATIONS. THE FORM OF THE EQUATIONS ARE:
    !*    A(I)*X(I-1) + B(I)*X(I) + C(I)*X(I+1) = D(I)

    !======================================================================!
    """
    AS=np.empty_like(AF)
    DS=np.empty_like(AF)
    XK=np.empty_like(AF)
    AS[-1] = AF[-1]/BF[-1]
    DS[-1] = DF[-1]/BF[-1]

    for I in range(1,L):
        X         = 1./(BF[L+1-I-2] - CF[L+1-I-2]*AS[L+2-I-2])
        AS[L+1-I-2] = AF[L+1-I-2]*X
        DS[L+1-I-2] = (DF[L+1-I-2]-CF[L+1-I-2]*DS[L+2-I-2])*X
 
    XK[0]=DS[0]
    for I in range(1,L):
        XKB   = XK[I-1]
        XK[I] = DS[I]-AS[I]*XKB
    return XK
    


# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Library of useful functions for users (only).

Functions here CANNOT be called into the library itself: importing this module in others would
lead to recursive import problems.
"""
import os
import numpy as np
from exo_k.ktable import Ktable
from exo_k.ktable5d import Ktable5d
from exo_k.xtable import Xtable
from exo_k.hires_spectrum import Hires_spectrum
from exo_k.util.cst import PI

def create_table(logpgrid=None, tgrid=None, xgrid=None, wngrid=None, wnedges=None,
        ggrid=None, weights=None, p_unit='Pa', kdata_unit='m^2/molecule',
        value=None, mol=None):
    """Create a new table with the required dimensions padded with zeros.

    The class of the object created will depend on the dimensions required. 

    Parameters
    ----------
        logpgrid: array
            Grid in log p
        tgrid: array
            Grid in T
        xgrid: array
            Grid in volume mixing ratio
        wngrid: array
            Grid wavenumber
        ggrid: array
            Grid g-space
        weights: array
            Grid of quadrature weights
        p_unit: str (optional)
            Unit of pressure (default Pa)
        kdata_unit: str (optional)
            Unit of cross section (default is m^2/molecule)

    Returns
    -------
        new_table: `Xtable`, `Ktable`, or `Ktable5d`
            A new table
    """
    if (ggrid is None) != (weights is None):
        raise RuntimeError("Both ggrid and weights should be specified or equal to None.")
    if (logpgrid is None) or (tgrid is None) or (wngrid is None):
        raise RuntimeError("You should at least specify logpgrid, tgrid, and wngrid.")
    if ggrid is not None:
        if xgrid is not None:
            new_table=Ktable5d()
            new_table.xgrid=np.array(xgrid, dtype=float)
            new_table.Nx=new_table.xgrid.size
        else:
            new_table=Ktable()
        new_table.ggrid=np.array(ggrid, dtype=float)
        new_table.weights=np.array(weights, dtype=float)
        if new_table.ggrid.size != new_table.weights.size:
            raise RuntimeError('ggrid and weights do not have the same length')
        new_table.Ng=new_table.ggrid.size
    else:
        new_table=Xtable()
    new_table.logpgrid=np.array(logpgrid, dtype=float)
    new_table.pgrid=10**new_table.logpgrid
    new_table.tgrid=np.array(tgrid, dtype=float)
    new_table.wns=np.array(wngrid, dtype=float)
    if wnedges is not None:
        new_table.wnedges=np.array(wnedges, dtype=float)
    else:
        new_table.wnedges=np.concatenate(  \
            ([new_table.wns[0]],(new_table.wns[:-1]+new_table.wns[1:])*0.5,[new_table.wns[-1]]))
    new_table.Np=new_table.pgrid.size
    new_table.Nt=new_table.tgrid.size
    new_table.Nw=new_table.wns.size
    sh=[new_table.Np,new_table.Nt,new_table.Nx,new_table.Nw,new_table.Ng]
    sh=[x for x in sh if x is not None]
    if value is None:
        new_table.kdata=np.zeros(sh, dtype=float)
    else:
        new_table.kdata=np.full(sh, value, dtype=float)
    new_table.p_unit=p_unit
    new_table.kdata_unit=kdata_unit
    new_table.mol=mol
    return new_table


def hires_to_ktable(filename_grid=None, xgrid=None, **kwargs):
    """Emulates :func:`exo_k.ktable.Ktable.hires_to_ktable`
    and :func:`exo_k.ktable5d.Ktable5d.hires_to_ktable` as functions and not methods
    so that the user can call them without creating a Ktable first.

    See those methods for details on the available arguments and options.
    """
    if filename_grid is not None:
        if len(np.array(filename_grid).shape) == 3 and xgrid is None:
            raise RuntimeError("""From the shape of filename_grid
                it seems you want have an xgrid dimension, but xgrid is None!
                """)    
    if xgrid is not None: # creates a Ktable5d
        if filename_grid is not None:
            if len(np.array(filename_grid).shape) != 3:
                raise RuntimeError("""You provided a xgrid of vmrs, 
                    but the shape of filename_grid is not 
                    of the type (Np, Nt, Nx)!
                    """)
        res=Ktable5d()
        res.hires_to_ktable(filename_grid=filename_grid, xgrid=xgrid, **kwargs)
    else:
        res=Ktable()
        res.hires_to_ktable(filename_grid=filename_grid, **kwargs)
    return res

def hires_to_xtable(**kwargs):
    """Emulates :func:`exo_k.xtable.Xtable.hires_to_xtable`
    as a function and not a method
    so that the user can call it without creating an Xtable first.

    See :func:`~exo_k.xtable.Xtable.hires_to_xtable` for details
    on the available arguments and options.
    """
    res=Xtable()
    res.hires_to_xtable(**kwargs)
    return res

def convert_kspectrum_to_hdf5(file_in, file_out=None, **kwargs):
    """Converts kspectrum like spectra to hdf5 format for speed and space.
    Helper function. Real work done in :class:`~exo_k.util.hires_spectrum.Hires_spectrum`
    __init__ funtion.

    Go there to see all the available arguments and options.

    Parameters
    ----------
        file_in: str
            Initial kspectrum filename.
        file_out: str
            Name of the final hdf5 file to be created. If not provided,
            'file_in.h5' will be used. 
    """
    tmp=Hires_spectrum(file_in, **kwargs)
    if file_out is None: file_out=file_in
    tmp.write_hdf5(file_out)

def create_fname_grid(base_string, logpgrid=None, tgrid=None, xgrid=None,
        p_kw=None, t_kw=None, x_kw=None):
    """Creates a grid of filenames from an array of pressures, temperatures (
    and vmr if there is a variable gas).

    Parameters
    ----------
        base_string: str
            Generic name of the spectra files with specific keywords to be replaced 
            by the relevant numerical values
        logpgrid: Array
            Grid in log(pressure/Pa) of the input
        tgrid: Array
            Grid in temperature of the input
        xgrid: Array
            Input grid in vmr of the variable gas
        
    .. warning::
        The result of this function is much more predictable 
        if the values in the above arrays are given as integers or directly strings. 
        If you want to use floats anyway, good luck. 

    Parameters
    ----------
        p_kw: str
        t_kw: str
        x_kw: str
            The pattern string that will be recognized as keywords between
            {} in base_string (See examples).

    Examples
    --------

        >>> logpgrid=[1,2]
        >>> tgrid=np.array([100.,150.,200.])
        >>> file_grid=exo_k.create_fname_grid('spectrum_CO2_1e{logp}Pa_{t}K.hdf5',
                  logpgrid=logpgrid,tgrid=tgrid,p_kw='logp',t_kw='t')                  
        array([['spectrum_CO2_1e1Pa_100K.hdf5', 'spectrum_CO2_1e1Pa_150K.hdf5',
        'spectrum_CO2_1e1Pa_200K.hdf5'],
        ['spectrum_CO2_1e2Pa_100K.hdf5', 'spectrum_CO2_1e2Pa_150K.hdf5',
        'spectrum_CO2_1e2Pa_200K.hdf5']], dtype='<U28')

    """
    logpgrid=np.array(logpgrid)
    tgrid=np.array(tgrid)
    res=[]
    if xgrid is None:
        for iP in range(logpgrid.size):
            for iT in range(tgrid.size):
                dict_opt={p_kw:str(logpgrid[iP]),t_kw:str(tgrid[iT])}
                fname=base_string.format(**dict_opt)
                res.append(fname)
        return np.array(res).reshape((logpgrid.size,tgrid.size))
    else:
        xgrid=np.array(xgrid)
        for iP in range(logpgrid.size):
            for iT in range(tgrid.size):
                for iX in range(xgrid.size):
                    dict_opt={p_kw:str(logpgrid[iP]), \
                        t_kw:str(tgrid[iT]),x_kw:str(xgrid[iX])}
                    fname=base_string.format(**dict_opt)
                    res.append(fname)
        return np.array(res).reshape((logpgrid.size,tgrid.size,xgrid.size))

def finalize_LMDZ_dir(corrkname, IRsize, VIsize):
    """Creates the right links for a LMDZ type directory to be read by the LMDZ generic GCM.

    You will need to create a proper Q.dat before using with the LMDZ GCM. 

    .. important::
        You must have run :func:`exo_k.ktable.Ktable.write_LMDZ` or
        :func:`exo_k.ktable5d.Ktable5d.write_LMDZ` for both of your IR and VI channels
        beforehand.

    Parameters
    ----------
        corrkname: str
            Path to the directory with the LMDZ ktable to finalize
        IRsize: int
            Number of IR spectral bins
        VIsize: int
            Number of VI spectral bins
    """
    newdir=os.path.join(corrkname,str(IRsize)+'x'+str(VIsize))
    try:
        os.mkdir(newdir)
    except FileExistsError:
        os.system('rm -rf '+newdir)
        os.mkdir(newdir)
    os.symlink('../IR'+str(IRsize)+'/corrk_gcm_IR.dat',os.path.join(newdir,'corrk_gcm_IR.dat'))
    os.symlink('../IR'+str(IRsize)+'/narrowbands_IR.in',os.path.join(newdir,'narrowbands_IR.in'))
    os.symlink('../VI'+str(VIsize)+'/corrk_gcm_VI.dat',os.path.join(newdir,'corrk_gcm_VI.dat'))
    os.symlink('../VI'+str(VIsize)+'/narrowbands_VI.in',os.path.join(newdir,'narrowbands_VI.in'))
    print('Everything went ok. Your ktable is in:',newdir)
    print("You'll probably need to add Q.dat before using it though!")

def mmr_to_number_density(mmr, gas_density, r_eff, condensate_density):
    """Converts a mass mixing ratio (mmr or q) in a number density of particles
    (in number per unit volume)

    Parameters
    ----------
        mmr: float or array
            Mass mixing ratio (in kg per kg of air)
        gas_density: float or array
            Density of the gas (in kg/m^3)
        r_eff: float or array
            Effective radius of the particles
        condensate_density: float or array
            Density of the constituent of the condensed particles (in kg/m^3)
    """
    particle_mass=4.*PI*r_eff**3*condensate_density/3.
    return mmr*gas_density/particle_mass
    


# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Contains classes for atmospheric profiles and their radiative properties.
That's where forward models are computed. 
"""
import numpy as np
import numba
from scipy.special import expn
import astropy.units as u
from numba.typed import List
from .gas_mix import Gas_mix
from .util.cst import N_A, PI, RGP, KBOLTZ, RSOL, RJUP, SIG_SB
from .util.radiation2 import Bnu_integral_num, Bnu, rad_prop_corrk, rad_prop_xsec,\
    Bnu_integral_array, path_integral_corrk, path_integral_xsec, dBnudT_array
from .two_stream import two_stream_toon2 as toon
from .two_stream import two_stream_lmdz as lmdz
from .two_stream import two_stream_crisp as crisp
from .util.interp import gauss_legendre
from .util.spectrum import Spectrum


class Atm_profile2(object):
    """A class defining an atmospheric PT profile with some global data
    (gravity, etc.)

    The derived class :class:`~exo_k.atm.Atm` handles
    radiative transfer calculations.

    """
    
    def __init__(self, composition={}, psurf=None, ptop=None, logplev=None, tlev=None,
            Tsurf=None, Tstrat=None, grav=None, Rp=None, Mgas=None, rcp=0.28, Nlev=20,
            **kwargs):
        """Initializes atmospheric profiles

        Parameters
        ----------
            composition: dict
                Keys are molecule names and values the vmr.
                Vmr can be arrays of size Nlev-1 (i.e. the number of layers).
            grav: float
                Planet surface gravity (gravity constant with altitude for now).
            Rp: float or Astropy.unit quantity
                Planet radius. If float, Jupiter radii are assumed.
            rcp: float
                Adiabatic lapse rate for the gas (R/cp)
            Mgas: float, optional
                Molar mass of the gas (kg/mol). If given, overrides the molar mass computed
                from composition.
        
        There are two ways to define the profile.
        Either define:

        * Nlev: int
          Number of level interfaces (Number of layers is Nlev-1)
        * psurf, Tsurf: float
          Surface pressure (Pa) and temperature 
        * ptop: float
          Pressure at the top of the model (Pa) 
        * Tstrat: float
          Stratospheric temperature        

        or:

        * logplev: array
        * tlev: array (same size)
          These will become the pressures (Pa) and temperatures at the level interfaces.
          This will be used to define the surface and top pressures.
          Nlev becomes the size of the arrays. 

        .. warning::
            Layers are counted from the top down (increasing pressure order).
            All methods follow the same convention.
        """
        self.gas_mix=Gas_mix(composition)
        self.rcp=rcp
        if logplev is None:
            self.Nlev=Nlev
            self.Nlay=Nlev-1
            self.logplev=np.linspace(np.log10(ptop),np.log10(psurf),num=self.Nlev)
            self.compute_pressure_levels()
            self.set_adiab_profile(Tsurf=Tsurf, Tstrat=Tstrat)
        else:
            self.set_logPT_profile(logplev, tlev)
        self.set_Rp(Rp)        
        self.set_grav(grav)
        self.set_Mgas(Mgas)

    def set_logPT_profile(self, log_plev, tlev):
        """Set the logP-T profile of the atmosphere with a new one

        Parameters
        ----------
            log_plev: numpy array
                Log pressure (in Pa) at the level surfaces
            tlev: numpy array (same size)
                temperature at the level surfaces.
        """
        self.logplev=np.array(log_plev, dtype=float)
        self.Nlev=log_plev.size
        self.Nlay=self.Nlev-1
        self.compute_pressure_levels()
        self.set_T_profile(tlev)

    def set_T_profile(self, tlev):
        """Reset the temperature profile without changing the pressure levels
        """
        tlev=np.array(tlev, dtype=float)
        if tlev.shape != self.logplev.shape:
            raise RuntimeError('tlev and log_plev should have the same size.')
        self.tlev=tlev
        self.tlay=(self.tlev[:-1]+self.tlev[1:])*0.5
        #self.tlay=self.tlev[:-1]
        self.gas_mix.set_logPT(logp_array=self.logplay, t_array=self.tlay)

    def compute_pressure_levels(self):
        """Computes various pressure related quantities
        """
        if self.logplev[0] >= self.logplev[-1]:
            print("""
            Atmospheres are modelled from the top down.
            All arrays should be ordered accordingly
            (first values correspond to top of atmosphere)""")
            raise RuntimeError('Pressure grid is in decreasing order!')
        self.plev=10**self.logplev
        self.psurf=self.plev[-1]
        #self.logplay=(self.logplev[:-1]+self.logplev[1:])*0.5
        #self.play=10**self.logplay
        self.play=(self.plev[:-1]+self.plev[1:])*0.5
        #self.play=self.plev[:-1]
        self.logplay=np.log10(self.play)
        self.dp_lay=np.concatenate([[self.plev[0]],self.play,[self.psurf]])
        self.dp_lay=np.diff(self.dp_lay)
        self.exner=(self.plev/self.psurf)**self.rcp

    def set_adiab_profile(self, Tsurf=None, Tstrat=None):
        """Initializes the logP-T atmospheric profile with an adiabat with index R/cp=rcp

        Parameters
        ----------
            Tsurf: float
                Surface temperature.
            Tstrat: float, optional
                Temperature of the stratosphere. If None is given,
                an isothermal atmosphere with T=Tsurf is returned.
        """
        if Tstrat is None: Tstrat=Tsurf
        self.tlev=Tsurf*self.exner
        self.tlev=np.where(self.tlev<Tstrat,Tstrat,self.tlev)
        self.tlay=(self.tlev[:-1]+self.tlev[1:])*0.5
        self.gas_mix.set_logPT(logp_array=self.logplay, t_array=self.tlay)

    def set_grav(self, grav=None):
        """Sets the surface gravity of the planet

        Parameters
        ----------
            grav: float
                surface gravity (m/s^2)
        """
        if grav is None: raise RuntimeError('A planet needs a gravity!')
        self.grav=grav
    
    def set_gas(self, composition_dict):
        """Sets the composition of the atmosphere

        Parameters
        ----------
            composition_dict: dictionary
                Keys are molecule names, and values are volume mixing ratios.
                A 'background' value means that the gas will be used to fill up to vmr=1
                If they do not add up to 1 and there is no background gas_mix,
                the rest of the gas_mix is considered transparent.
    
            compute_col_dens: boolean, optional
                If True, the column density per layer of the atmosphere is recomputed.
                This si mostly to save time when we know this will be done later on.
        """
        self.gas_mix.set_composition(composition_dict)
        self.set_Mgas()

    def set_Mgas(self, Mgas=None):
        """Sets the mean molar mass of the atmosphere.

        Parameters
        ----------
            Mgas: float or array of size Nlay
                Mean molar mass (kg/mol).
                If None is give, the mmm is computed from the composition.
        """
        if Mgas is not None:
            self.Mgas=Mgas
        else:
            self.Mgas=self.gas_mix.molar_mass()

    def set_rcp(self,rcp):
        """Sets the adiabatic index of the atmosphere

        Parameters
        ----------
            rcp: float
                R/c_p
        """
        self.rcp=rcp

    def set_Rp(self, Rp):
        """Sets the radius of the planet

        Parameters
        ----------
            Rp: float
                radius of the planet (m)
        """
        if Rp is None:
            self.Rp = None
            return
        if isinstance(Rp,u.quantity.Quantity):
            self.Rp=Rp.to(u.m).value
        else:
            self.Rp=Rp*RJUP

    def set_Rstar(self, Rstar):
        """Sets the radius of the star

        Parameters
        ----------
            Rstar: float
                radius of the star (m)
        """
        if Rstar is None:
            self.Rstar = None
            return
        if isinstance(Rstar,u.quantity.Quantity):
            self.Rstar=Rstar.to(u.m).value
        else:
            self.Rstar=Rstar*RSOL

    def compute_density(self):
        """Computes the number density (m^-3) profile of the atmosphere
        """
        self.density=self.play/(KBOLTZ*self.tlay)

    def compute_layer_col_density(self):
        """Computes the column number density (molecules/m^-2) per layer of the atmosphere
        """
        self.dmass=(self.plev[1:]-self.plev[:-1])/self.grav
        # grav term above should include the altitude effect. 
        #self.dmass[0]=self.plev[1]/self.grav
        # above is just a test extending the atm up to p=0
        if self.Rp is not None:
            self.compute_altitudes()
            self.dmass=self.dmass*(1.+self.zlay/self.Rp)**2
        self.dcol_density=self.dmass*N_A/(self.Mgas)

    def compute_altitudes(self):
        """Compute altitudes of the level surfaces (zlev) and mid layers (zlay).
        """
        H=RGP*self.tlay/(self.grav*self.Mgas)
        dlnP=np.diff(self.logplev)*np.log(10.)
        if self.Rp is None:
            self.dz=H*dlnP
            self.zlay=np.cumsum(self.dz[::-1])
            self.zlev=np.concatenate(([0.],self.zlay))[::-1]
            self.zlay-=0.5*self.dz[::-1]
            self.zlay=self.zlay[::-1]
        else:
            self.zlev=np.zeros_like(self.tlev)
            for i in range(H.size)[::-1]:
                z1=self.zlev[i+1]
                H1=H[i]
                dlnp=dlnP[i]
                self.zlev[i]=z1+( (H1 * (self.Rp + z1)**2 * dlnp) \
                    / (self.Rp**2 + H1 * self.Rp * dlnp + H1 * z1 * dlnp) )
            self.zlay=(self.zlev[:-1]+self.zlev[1:])*0.5
            # the last line is not completely consistent. The integration
            # should be done between layer bottom and mid layer assuming
            # an average T between Tlay and Tlev.
        
    def compute_area(self):
        """Computes the area of the annulus covered by each layer in a transit setup. 
        """
        self.area=PI*(self.Rp+self.zlev[:-1])**2
        self.area[:-1]-=self.area[1:]
        self.area[-1]-=PI*self.Rp**2

    def compute_tangent_path(self):
        """Computes a triangular array of the tangent path length (in m) spent in each layer.
        self.tangent_path[ilay][jlay] is the length that the ray that is tangent to the ilay layer 
        spends in the jlay>=ilay layer (accounting for a factor of 2 due to symmetry)
        """
        if self.Rp is None: raise RuntimeError('Planetary radius should be set')
        self.compute_altitudes()
        self.tangent_path=List()
        # List() is a new numba.typed list to comply with new numba evolution after v0.50
        for ilay in range(self.Nlay):
            z0square=(self.Rp+self.zlay[ilay])**2
            dl=np.sqrt((self.Rp+self.zlev[:ilay+1])**2-z0square)
            dl[:-1]-=dl[1:]
            self.tangent_path.append(2.*dl)

    def __repr__(self):
        """Method to output header
        """
        output="""
    gravity (m/s^2) : {grav}
    Planet Radius(m): {rad}
    Ptop (Pa)       : {ptop}
    Psurf (Pa)      : {psurf}
    composition     :
        {comp}""".format(grav=self.grav, rad=self.Rp, comp=self.gas_mix,
            ptop=self.plev[0], psurf=self.plev[-1])
        return output



class Atm2(Atm_profile2):
    """Class based on Atm_profile that handles radiative trasnfer calculations.

    Radiative data are accessed through the :any:`gas_mix.Gas_mix` class.
    """

    def __init__(self, k_database=None, cia_database=None,
        wn_range=None, wl_range=None, **kwargs):
        """Initialization method that calls Atm_Profile().__init__() and links
        to Kdatabase and other radiative data. 
        """
        super().__init__(**kwargs)
        self.set_k_database(k_database)
        self.set_cia_database(cia_database)
        self.set_spectral_range(wn_range=wn_range, wl_range=wl_range)
        self.flux_net_nu=None
        self.kernel=None
        self.tlev_kernel=self.tlev

    def set_k_database(self, k_database=None):
        """Change the radiative database used by the
        :class:`Gas_mix` object handling opacities inside
        :class:`Atm`.

        See :any:`gas_mix.Gas_mix.set_k_database` for details.

        Parameters
        ----------
            k_database: :class:`Kdatabase` object
                New Kdatabase to use.
        """
        self.gas_mix.set_k_database(k_database=k_database)
        self.kdatabase=self.gas_mix.kdatabase
        self.Ng=self.gas_mix.Ng
        # to know whether we are dealing with corr-k or not and access some attributes. 

    def set_cia_database(self, cia_database=None):
        """Change the CIA database used by the
        :class:`Gas_mix` object handling opacities inside
        :class:`Atm`.

        See :any:`gas_mix.Gas_mix.set_cia_database` for details.

        Parameters
        ----------
            cia_database: :class:`CIAdatabase` object
                New CIAdatabase to use.
        """
        self.gas_mix.set_cia_database(cia_database=cia_database)

    def set_spectral_range(self, wn_range=None, wl_range=None):
        """Sets the spectral range in which computations will be done by specifying
        either the wavenumber (in cm^-1) or the wavelength (in micron) range.

        See :any:`gas_mix.Gas_mix.set_spectral_range` for details.
        """
        self.gas_mix.set_spectral_range(wn_range=wn_range, wl_range=wl_range)

    def spectral_integration(self, spectral_var):
        """Spectrally integrate an array, taking care of whether
        we are dealing with corr-k or xsec data.

        Parameters
        ----------
            spectral_var: array
                array to integrate

        Returns
        -------
            var: array
                array integrated over wavenumber (and g-space if relevant)
        """
        if self.Ng is None:
            var=np.sum(spectral_var*self.dwnedges,axis=-1)
        else:
            var=np.sum(np.sum(spectral_var*self.weights,axis=-1)*self.dwnedges,axis=-1)
        return var

    def opacity(self, rayleigh = False, **kwargs):
        """Computes the opacity of each of the layers.

        See :any:`gas_mix.Gas_mix.cross_section` for details.
        """
        self.kdata = self.gas_mix.cross_section(rayleigh=rayleigh, **kwargs)
        if rayleigh: self.kdata_scat=self.gas_mix.kdata_scat
        self.Nw=self.gas_mix.Nw
        self.wns=self.gas_mix.wns
        self.wnedges=self.gas_mix.wnedges
        self.dwnedges=self.gas_mix.dwnedges

    def source_function(self, integral=True, source=True):
        """Compute the blackbody source function (Pi*Bnu) for each layer of the atmosphere.

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 
            source: boolean, optional
                If False, the source function is put to 0 (for solar absorption calculations)
        """
        if source:
            if integral:
                #JL2020 What is below is slightly faster for very large resolutions
                #piBatm=np.empty((self.Nlev,self.Nw))
                #for ii in range(self.Nlev):
                #    piBatm[ii]=PI*Bnu_integral_num(self.wnedges,self.tlev[ii])/dw
                #JL2020 What is below is much faster for moderate to low resolutions
                piBatm=PI*Bnu_integral_array(self.wnedges,self.tlev,self.Nw,self.Nlev) \
                    /self.dwnedges
            else:
                piBatm=PI*Bnu(self.wns[None,:],self.tlev[:,None])
        else:
            piBatm=np.zeros((self.Nlev,self.Nw))
        return piBatm

    def setup_emission_caculation(self, mu_eff=0.5, rayleigh=False, integral=True,
            source=True, **kwargs):
        """Computes all necessary quantities for emission calculations
        (opacity, source, etc.)
        """
        self.opacity(rayleigh=rayleigh, **kwargs)
        self.piBatm = self.source_function(integral=integral, source=source)
        self.compute_layer_col_density()
        if self.Ng is None:
            self.tau,self.dtau=rad_prop_xsec(self.dcol_density,self.kdata,mu_eff)
        else:
            self.tau,self.dtau=rad_prop_corrk(self.dcol_density,self.kdata,mu_eff)
            self.weights=self.kdatabase.weights

    def emission_spectrum(self, integral=True, mu0=0.5, mu_quad_order=None, **kwargs):
        """Returns the emission flux at the top of the atmosphere (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 

        Other Parameters
        ----------------
            mu0: float
                Cosine of the quadrature angle use to compute output flux
            mu_quad_order: int
                If an integer is given, the emission intensity is computed
                for a number of angles and integrated following a gauss legendre
                quadrature rule of order `mu_quad_order`.

        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        if mu_quad_order is not None:
            # if we want quadrature, use the more general method.
            return self.emission_spectrum_quad(integral=integral,
                mu_quad_order=mu_quad_order, **kwargs)

        try:
            self.setup_emission_caculation(mu_eff=mu0, rayleigh=False, integral=integral, **kwargs)
        except TypeError:
            raise RuntimeError("""Cannot use rayleigh option with emission_spectrum.
            Probably meant to use emission_spectrum_2stream.
            """)
        # self.tau and self.dtau include the 1/mu0 factor.
        expdtau=np.exp(-self.dtau)
        expdtauminone=np.where(self.dtau<1.e-13,-self.dtau,expdtau-1.)
        # careful: due to numerical limitations, 
        # the limited development of Exp(-dtau)-1 needs to be used for small values of dtau
        exptau=np.exp(-self.tau)
        if self.Ng is None:
            timesBatmTop=(1.+expdtauminone/self.dtau)*exptau[:-1]
            timesBatmBottom=(-expdtau-expdtauminone/self.dtau)*exptau[:-1]
            timesBatmBottom[-1]+=exptau[-1]
        else:
            timesBatmTop=np.sum((1.+expdtauminone/self.dtau)*exptau[:-1]*self.weights,axis=-1)
            timesBatmBottom=np.sum((-expdtau-expdtauminone/self.dtau)*exptau[:-1] \
                *self.weights,axis=-1)
            timesBatmBottom[-1]+=np.sum(exptau[-1]*self.weights,axis=-1)
        IpTop=np.sum(self.piBatm[:-1]*timesBatmTop+self.piBatm[1:]*timesBatmBottom,axis=0)

        return Spectrum(IpTop,self.wns,self.wnedges)

    def emission_spectrum_quad(self, integral=True, mu_quad_order=3, **kwargs):
        """Returns the emission flux at the top of the atmosphere (in W/m^2/cm^-1)
        using gauss legendre qudrature of order `mu_quad_order`

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 
                  
        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        self.setup_emission_caculation(mu_eff=1., rayleigh=False, integral=integral, **kwargs)
        # angle effect dealt with later

        IpTop=np.zeros(self.kdata.shape[1])
        mu_w, mu_a, _ = gauss_legendre(mu_quad_order)
        mu_w = mu_w * mu_a * 2.# takes care of the mu factor in last integral => int(mu I d mu)
                               # Factor 2 takes care of the fact that the source function is pi*Batm
                               # but we want 2*Pi*Batm
        for ii, mu0 in enumerate(mu_a):
            tau=self.tau/mu0
            dtau=self.dtau/mu0
            expdtau=np.exp(-dtau)
            expdtauminone=np.where(dtau<1.e-13,-dtau,expdtau-1.)
            exptau=np.exp(-tau)
            if self.Ng is None:
                timesBatmTop=(-expdtau-expdtauminone/dtau)*exptau[:-1]
                timesBatmBottom=(1.+expdtauminone/dtau)*exptau[:-1]
                timesBatmBottom[-1]=timesBatmBottom[-1]+exptau[-1]
            else:
                timesBatmTop=np.sum((-expdtau-expdtauminone/dtau)*exptau[:-1] \
                    *self.weights,axis=-1)
                timesBatmBottom=np.sum((1.+expdtauminone/dtau)*exptau[:-1] \
                    *self.weights,axis=-1)
                timesBatmBottom[-1]=timesBatmBottom[-1]+np.sum(exptau[-1]*self.weights,axis=-1)
            IpTop+=np.sum(self.piBatm[:-1]*timesBatmTop+self.piBatm[1:]*timesBatmBottom,axis=0) \
                *mu_w[ii]

        return Spectrum(IpTop,self.wns,self.wnedges)

    def emission_spectrum_2stream(self, integral=True, mu0=0.5,
            method='toon', dtau_min=1.e-10, mid_layer=False, rayleigh=False,
            flux_top_dw=None, source=True, compute_kernel=False, **kwargs):
        """Returns the emission flux at the top of the atmosphere (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 

        Other Parameters
        ----------------
            mu0: float
                Cosine of the quadrature angle use to compute output flux
            dtau_min: float
                If the optical depth in a layer is smaller than dtau_min,
                dtau_min is used in that layer instead. Important as too
                transparent layers can cause important numerical rounding errors.

        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        self.setup_emission_caculation(mu_eff=1., rayleigh=rayleigh, integral=integral,
            source=source, **kwargs)
        # mu_eff=1. because the mu effect is taken into account in solve_2stream_nu
                 # we must compute the vertical optical depth here.
        self.dtau=np.where(self.dtau<dtau_min,dtau_min,self.dtau)

        module_to_use=globals()[method]
        # globals()[method] converts the method string into a module name
        #  if the module has been loaded
        if self.Ng is None:
            solve_2stream_nu=module_to_use.solve_2stream_nu_xsec
        else:
            solve_2stream_nu=module_to_use.solve_2stream_nu_corrk

        if rayleigh:
            if self.Ng is None:
                self.single_scat_albedo = self.kdata_scat / self.kdata
            else:
                self.single_scat_albedo = self.kdata_scat[:,:,None] / self.kdata
        else:
            self.single_scat_albedo = np.zeros_like(self.dtau)
        self.single_scat_albedo=np.core.umath.minimum(self.single_scat_albedo,0.9999999999999)
        self.asym_param = np.zeros_like(self.dtau)

        if flux_top_dw is None:
            self.flux_top_dw_nu = np.zeros((self.Nw))
        else:
            Tstar=5700.
            self.flux_top_dw_nu = Bnu_integral_num(self.wnedges,Tstar)
            self.flux_top_dw_nu = self.flux_top_dw_nu * flux_top_dw \
                / (np.sum(self.flux_top_dw_nu)*self.dwnedges)
            #self.flux_top_dw_nu=Spectrum(flux_top_dw_nu,self.wns,self.wnedges)

        if method == 'crisp':
            self.flux_up_nu, self.flux_down_nu, self.flux_net_nu, kernel_nu = \
                solve_2stream_nu(self.piBatm, self.dtau,
                self.single_scat_albedo, self.asym_param, self.flux_top_dw_nu, mu0 = mu0)
            if compute_kernel:
                dB = PI * dBnudT_array(self.wns, self.tlev, self.Nw, self.Nlev)
                if self.Ng is None:
                    self.kernel=np.sum(kernel_nu*dB*self.dwnedges,axis=2)
                else:
                    self.kernel=np.sum(np.sum(kernel_nu*self.weights,axis=3) \
                        *dB*self.dwnedges,axis=2)
                self.kernel[:,:-1]=(self.kernel[:,:-1]-self.kernel[:,1:])
                self.kernel*=self.grav/self.dp_lay
        else:
            self.flux_up_nu, self.flux_down_nu, self.flux_net_nu = \
                solve_2stream_nu(self.piBatm, self.dtau, self.single_scat_albedo, self.asym_param,
                    self.flux_top_dw_nu, mu0 = mu0, mid_layer=mid_layer)
            if compute_kernel: self.compute_kernel(solve_2stream_nu, mu0=mu0, mid_layer=mid_layer,
                        per_unit_mass=True, integral=True, **kwargs)

        if self.Ng is None:
            return Spectrum(self.flux_up_nu[0],self.wns,self.wnedges)
        else:
            return Spectrum(np.sum(self.flux_up_nu[0]*self.weights,axis=1),self.wns,self.wnedges)

    def compute_kernel(self, solve_2stream_nu, epsilon=0.01, mid_layer=False, mu0 = 0.5,
            per_unit_mass=True, integral=True, **kwargs):
        """Compute the Jacobian matrix d Heating[lev=i] / d T[lev=j]
        """
        net=self.spectral_integration(self.flux_net_nu)
        self.kernel=np.empty((self.Nlev,self.Nlev))
        #dB = PI * dBnudT_array(self.wns, self.tlev, self.Nw, self.Nlev)
        tlev=self.tlev
        dT = epsilon*tlev
        self.tlev = tlev + dT
        newpiBatm = self.source_function(integral=integral)

        #newpiBatm=PI*Bnu_integral_array(self.wnedges,self.tlev+dT,self.Nw,self.Nlev) \
        #            /self.dwnedges

        for ilev in range(self.Nlev):
            pibatm = np.copy(self.piBatm)
            #dT = epsilon*self.tlev[ilev]
            #pibatm[ilev] += dB[ilev]*dT
            pibatm[ilev] = newpiBatm[ilev]
            _, _, flux_net_tmp = \
                solve_2stream_nu(pibatm, self.dtau, self.single_scat_albedo, self.asym_param,
                    self.flux_top_dw_nu, mu0 = mu0, mid_layer=mid_layer)
            net_tmp = self.spectral_integration(flux_net_tmp)
            #self.kernel[ilev]=(net_tmp-net)/dT
            self.kernel[ilev]=(net-net_tmp)/dT[ilev]
        self.kernel[:,:-1]-=self.kernel[:,1:]
        self.tlev=tlev
        if per_unit_mass: self.kernel*=self.grav/self.dp_lay

    def flux_divergence(self, internal_flux=0., per_unit_mass = True):
        """Computes the divergence of the net flux in the layers
        (used to compute heating rates).

        :func:`emission_spectrum_2stream` needs to be ran first.

        Parameters
        ----------
            internal_flux: float
                flux coming from below in W/m^2
            per_unit_mass: bool
                If True, the heating rates are normalized by the
                mass of each layer (result in W/kg).

        Returns
        -------
            net: array
                Net fluxes at level surfaces
            H: array
                Heating rate in each layer (Difference of the net fluxes). Positive means heating.
                The last value is the net flux impinging on the surface + the internal flux.
        """
        if self.flux_net_nu is None:
            raise RuntimeError('should have ran emission_spectrum_2stream.')
        net=self.spectral_integration(self.flux_net_nu)
        H=-np.copy(net)
        H[:-1]-=H[1:]
        H[-1]+=internal_flux
        if per_unit_mass: H*=self.grav/self.dp_lay
        return net, H

    def bolometric_fluxes(self, internal_flux = 0., per_unit_mass = True):
        """Computes the bolometric fluxes at levels and the divergence of the net flux in the layers
        (used to compute heating rates).

        :func:`emission_spectrum_2stream` needs to be ran first.

        Parameters
        ----------
            internal_flux: float
                flux coming from below in W/m^2
            per_unit_mass: bool
                If True, the heating rates are normalized by the
                mass of each layer (result in W/kg).

        Returns
        -------
            up: array
                Upward fluxes at level surfaces
            dw: array
                Downward fluxes at level surfaces
            net: array
                Net fluxes at level surfaces
            H: array
                Heating rate in each layer (Difference of the net fluxes). Positive means heating.
                The last value is the net flux impinging on the surface + the internal flux.
        """
        net, H = self.flux_divergence(internal_flux=internal_flux, per_unit_mass = per_unit_mass)
        up=self.spectral_integration(self.flux_up_nu)
        dw=self.spectral_integration(self.flux_down_nu)
        return up, dw, net, H

    def exp_minus_tau(self):
        """Sums Exp(-tau) over gauss points
        """
        weights=self.kdatabase.weights
        return np.sum(np.exp(-self.tau[1:])*weights,axis=2)

    def exp_minus_tau_g(self, g_index):
        """Sums Exp(-tau) over gauss points
        """
        return np.exp(-self.tau[1:,:,g_index])

    def surf_bb(self, integral=True):
        """Computes the surface black body flux (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 
        Returns
        -------
            Spectrum object
                Spectral flux at the surface (in W/m^2/cm^-1)
        """
        if integral:
            piBatm=PI*Bnu_integral_num(self.wnedges,self.tlev[-1])/np.diff(self.wnedges)
        else:
            piBatm=PI*Bnu(self.wns[:],self.tlev[-1])
        return Spectrum(piBatm,self.wns,self.wnedges)

    def top_bb(self, integral=True):
        """Computes the top of atmosphere black body flux (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                If true, the black body is integrated within each wavenumber bin.
                If not, only the central value is used.
                    False is faster and should be ok for small bins,
                    but True is the correct version. 
        Returns
        -------
            Spectrum object
                Spectral flux of a bb at the temperature at the top of atmosphere (in W/m^2/cm^-1)
        """
        if integral:
            piBatm=PI*Bnu_integral_num(self.wnedges,self.tlev[0])/np.diff(self.wnedges)
        else:
            piBatm=PI*Bnu(self.wns[:],self.tlev[0])
        return Spectrum(piBatm,self.wns,self.wnedges)

    def transmittance_profile(self, **kwargs):
        """Computes the transmittance profile of an atmosphere,
        i.e. Exp(-tau) for each layer of the model.
        Real work done in the numbafied function path_integral_corrk/xsec
        depending on the type of data.
        """
        self.opacity(**kwargs)
        self.compute_tangent_path()
        self.compute_density()
        if self.Ng is not None:
            self.weights=self.kdatabase.weights
            transmittance=path_integral_corrk( \
                self.Nlay,self.Nw,self.Ng,self.tangent_path,self.density,self.kdata,self.weights)
        else:
            transmittance=path_integral_xsec( \
                self.Nlay,self.Nw,self.tangent_path,self.density,self.kdata)
        return transmittance

    def transmission_spectrum(self, normalized=False, Rstar=None, **kwargs):
        r"""Computes the transmission spectrum of the atmosphere.
        In general (see options below), the code returns the transit depth:

        .. math::
            \delta_\nu=(\pi R_p^2+\alpha_\nu)/(\pi R_{star}^2),

        where

        .. math::
          \alpha_\nu=2 \pi \int_0^{z_{max}} (R_p+z)*(1-e^{-\tau_\nu(z)) d z.
        
        Parameters
        ----------
            Rstar: float, optional
                Radius of the host star. Does not need to be given here if
                as already been specified as an attribute of the self.Atm object.
                If specified, the result is the transit depth:

                .. math::
                  \delta_\nu=(\pi R_p^2+\alpha_\nu)/(\pi R_{star}^2).

            normalized: boolean, optional
                Used only if self.Rstar and Rstar are None:

                * If True,
                  the result is normalized to the planetary radius:

                  .. math::
                    \delta_\nu=1+\frac{\alpha_\nu}{\pi R_p^2}.
                * If False,
                                    
                  .. math::
                    \delta_\nu=\pi R_p^2+\alpha_\nu.

        Returns
        -------
            array
                The transit spectrum (see above for normalization options).
        """
        self.set_Rstar(Rstar)
        transmittance=self.transmittance_profile(**kwargs)
        self.compute_area()
        res=Spectrum((np.dot(self.area,(1.-transmittance))),self.wns,self.wnedges)
        if self.Rstar is not None:
            return (res+(PI*self.Rp**2))/(PI*self.Rstar**2)
        elif normalized:
            return res/(PI*self.Rp**2)+1
        else:
            return res+(PI*self.Rp**2)

    def heating_rate(self, Fin=1., Tstar=5570., szangle=60., **kwargs):
        """Computes the heating rate in the atmosphere

        Parameters
        ----------
            Fin: float
                Bolometric stellar flux at the top of atmopshere (W/m^2).
            Tstar: float
                Stellar temperature
            szangle: float
                Solar zenith angle

        Returns
        -------
            array
                Heating rate in each atmospheric layer (K/s).  
        """
        mu0=np.cos(szangle*PI/180.)
        self.opacity(rayleigh=False, **kwargs)
        Fstar=Fin*PI*Bnu_integral_array(self.wnedges,List([Tstar]),self.Nw,1)/(SIG_SB*Tstar**4)
        self.compute_layer_col_density()
        if self.Ng is None:
            self.tau, _ =rad_prop_xsec(self.dcol_density,self.kdata,mu0)
            #the second returned variable is ignored
        else:
            self.tau, _ =rad_prop_corrk(self.dcol_density,self.kdata,mu0)
            self.weights=self.kdatabase.weights
        exptau=np.exp(-self.tau)
        if self.Ng is None:
            tmp_heat_rate=Fstar[0,:]*exptau
            tmp_heat_rate[:-1]-=tmp_heat_rate[1:]
            heat_rate=tmp_heat_rate[:-1]
        else:
            tmp_heat_rate=Fstar[0,:,None]*exptau
            tmp_heat_rate[:-1]-=tmp_heat_rate[1:]
            heat_rate=np.sum(tmp_heat_rate[:-1]*self.weights,axis=2)
        heat_rate=np.sum(heat_rate,axis=1)
        heat_rate=heat_rate*N_A*self.rcp/(self.dcol_density*RGP)
        return heat_rate

    def __repr__(self):
        """Method to output header
        """
        output=super().__repr__()
        output+="""
    k_database      :
        {kdatab}
    cia_database    :
        {cdatab}""".format(kdatab=self.kdatabase, cdatab=self.gas_mix.cia_database)
        if self.gas_mix._wn_range is not None:
            output+='    wn range        : '+ self.gas_mix._wn_range +'\n'

        return output


############### unvalidated functions #########################

    def emission_spectrum_exp_integral(self, integral=True, **kwargs):
        """Computes the emission flux at the top of the atmosphere (in W/m^2/cm^-1)

        .. warning::
            This method uses a formulation with exponential integrals that
            is not yet perceived as accurate enough by the author. It is left for testing only. 

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 

        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        #if self.dtau is None:
        #    self.rad_prop(**kwargs)
        self.opacity(rayleigh=False, **kwargs)
        if integral:
            piBatm=2*PI*Bnu_integral_array(self.wnedges,self.tlev,self.Nw,self.Nlev) \
                /np.diff(self.wnedges)
        else:
            piBatm=2*PI*Bnu(self.wns[None,:],self.tlev[:,None])
        mu_zenith=1.
        self.compute_layer_col_density()
        if self.Ng is None:
            self.tau,self.dtau=rad_prop_xsec(self.dcol_density, self.kdata, mu_zenith)
        else:
            self.tau,self.dtau=rad_prop_corrk(self.dcol_density, self.kdata, mu_zenith)
            self.weights=self.kdatabase.weights
        factor=expn(2, self.tau[1:])*self.dtau
        if self.Ng is None:
            timesBatmTop=factor*piBatm[:-1]
            surf_contrib=piBatm[-1]*expn(3, self.tau[-1])
        else:
            timesBatmTop=piBatm[:-1]*np.sum(factor*self.weights,axis=2)
            surf_contrib=piBatm[-1]*np.sum(expn(3, self.tau[-1])*self.weights,axis=1)
        IpTop=np.sum(timesBatmTop,axis=0)+surf_contrib
        return Spectrum(IpTop, self.wns, self.wnedges)

@numba.jit(nopython=True,fastmath=True)
def convadj(timestep, Nlev, tlev, exner, dp_lay, verbose = False):
    r"""Computes the heating rates needed to adjust unstable regions 
    of a given atmosphere to a convectively neutral T profile on
    a given timestep.

    .. important::
        The *layers* here are not the same as the *layers* in the radiative transfer!

        In the radiative transfer, a layer is the volume between two pressure level boundaries
        (`plev`).

        Here, a layer is centered on a pressure level (`plev`) with its temperature `tlev`.
        Therefore, `dp_lay[0]=play[0]-plev[0]`, `dp_lay[i]=play[i]-play[i-1]`, and
        `dp_lay[-1]=psurf-play[-1]`.
    
    Parameters
    ----------
        timestep: float
            Duration of the adjustment in seconds.
        Nlev: int
            Number of atmospheric layers
        tlev: array
            Temperatures of the atmospheric layers
        exner: array
            Exner function computed at the layer centers ((p/psurf)**rcp)

            .. math::
              \Pi=(p / p_{s})^{R/c_p}

        dp_lay: array
            Pressure difference between the bottom and top of each layer

    Returns
    -------
        array
            Heating rate in each atmospheric layer (K/s).  
    """
    epsilon=-1.e-5
    theta_lev=tlev/exner
    new_theta_lev=np.copy(theta_lev)
    dsig=dp_lay
    sdsig=dsig*exner
    H_conv=np.zeros(Nlev)
    n_iter=0
    while True:
        conv=np.nonzero(new_theta_lev[:-1]-new_theta_lev[1:]<epsilon)[0]
        # find convective layers
        N_conv=conv.size
        if N_conv==0: # no more convective regions, can exit
            return H_conv
        i_conv=0
        i_top=conv[i_conv] #upper unstable layer
        while i_conv<N_conv-1: #search from the top of the 1st unstable layer for its bottom
            if conv[i_conv+1]==conv[i_conv]+1:
                i_conv+=1
                continue
            else:
                break
        i_bot=conv[i_conv]+1
        mass_conv=0.
        intexner=0.
        theta_mean=0.
        for ii in range(i_top,i_bot+1): # compute new mean potential temperature
            intexner+=sdsig[ii]
            mass_conv+=dsig[ii]
            theta_mean+=sdsig[ii] * (theta_lev[ii] - theta_mean) / intexner
        while i_top>0:#look for newly unstable layers above
            if theta_lev[i_top-1]<theta_mean:
                i_top-=1
                intexner+=sdsig[i_top]
                mass_conv+=dsig[i_top]
                theta_mean+=sdsig[i_top] * (theta_lev[i_top] - theta_mean) / intexner
            else:
                break
        while i_bot<Nlev-1: #look for newly unstable layers below
            if theta_lev[i_bot+1]>theta_mean:
                i_bot+=1
                intexner+=sdsig[i_bot]
                mass_conv+=dsig[i_bot]
                theta_mean+=sdsig[i_bot] * (theta_lev[i_bot] - theta_mean) / intexner
            else:
                break
        # compute heating and adjust before looking for a new potential unstable layer
        H_conv[i_top:i_bot+1]=(theta_mean-new_theta_lev[i_top:i_bot+1]) \
            *exner[i_top:i_bot+1]/timestep
        new_theta_lev[i_top:i_bot+1]=theta_mean
        n_iter+=1
        if n_iter>Nlev+1:
            if verbose : print('oops, went crazy in convadj')
            break
    return H_conv



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
from scipy.interpolate import RectBivariateSpline
import pandas as pd
from .util.interp import bilinear_interpolation_array, interp_ind_weights

class EquChemTable(object):
    """Class to load and interpolate chemistry data.
    """
    
    def __init__(self, filename=None, remove_zeros=False):
        """Initializes chemical composition tables from various init files

        Parameters
        ----------
            filename : str
                Name of the input file
        """
        self.filename=filename
        self.tgrid=None
        self.pgrid=None
        self.logpgrid=None
        self.p_unit=None
        self.Nt=None
        self.Np=None
        self.molecules=None
        self.vol_mix_ratio={}
        self.logx_interp={}
        if self.filename.lower().endswith('.in'):
            self.read_composition_in(filename=self.filename, remove_zeros=remove_zeros)
        elif self.filename.lower().endswith('.dat'):
            self.read_composition_dat(filename=self.filename, remove_zeros=remove_zeros)
        self.setup_interpolation()

    def read_composition_in(self, filename=None, remove_zeros=False, skiprows=7):
        """Initializes chemical composition tables from composition.in files

        Parameters
        ----------
            filename : str
                Name of the input file
            skiprows : int, optional
                Number of lines to skip in the file
        """
        data = np.loadtxt(filename,skiprows=skiprows,unpack=True)
        self.Nt=np.where(data[2,1:]==data[2,0])[0][0]+1
        self.tgrid=data[2,0:self.Nt]
        #print('T grid:',self.tgrid)
        self.Np=data[0].size//self.Nt
        self.pgrid=data[1,::self.Nt]/0.986923267*1.e5
        self.p_unit='Pa'
        #print('P grid:',self.pgrid)
        #print('Np, Nt:',self.Np, self.Nt)
        self.logpgrid=np.log10(self.pgrid)
        with open(filename, 'r') as file:
            last_line=file.readline()
            while True:
                new_line=file.readline()
                if new_line.split()[0]=='z':
                    break
                last_line=new_line
        molecules=last_line.split()
        #print(molecules)
        for ii, mol in enumerate(molecules):
            self.vol_mix_ratio[mol]=data[ii+3].reshape((self.Np,self.Nt))
        if remove_zeros:
            self.remove_zeros()

    def read_composition_dat(self, filename=None, remove_zeros=False, skiprows=1):
        """Initializes chemical composition tables from composition.in files

        Parameters
        ----------
            filename : str
                Name of the input file
            skiprows : int, optional
                Number of lines to skip in the file
        """
        with open(filename, 'r') as file:
            molecules=file.readline().split()[2:]
        molecules[0]=molecules[0].replace('[mol/mol]','')
        data = np.loadtxt(filename,skiprows=skiprows,unpack=True)
        self.pgrid=np.sort(np.array(list(set(data[0]))))
        self.logpgrid=np.log10(self.pgrid)
        self.tgrid=np.sort(np.array(list(set(data[1]))))
        self.Np=self.pgrid.size
        self.Nt=self.tgrid.size
        for ii, mol in enumerate(molecules):
            self.vol_mix_ratio[mol]=data[ii+2].reshape((self.Np,self.Nt))[:,-1::-1]
        if remove_zeros:
            self.remove_zeros()

    def remove_zeros(self,deltalog_min_value=30.):
        """Finds zeros in the chem data and set them to (10.^-deltalog_min_value)
        times the minimum positive value
        in the table. This is to be able to work in logspace. 
        """
        for mol,x_ratio in self.vol_mix_ratio.items():
            mask = np.zeros((self.Np,self.Nt),dtype=bool)
            mask[np.nonzero(x_ratio)] = True
            minvalue=np.amin(x_ratio[mask])
            self.vol_mix_ratio[mol][~mask]=minvalue/(10.**deltalog_min_value)

    def setup_interpolation(self):
        """Creates interpolating functions to be called later on.
        """
        for mol,x_ratio in self.vol_mix_ratio.items():
            self.logx_interp[mol]=RectBivariateSpline(self.logpgrid,self.tgrid,np.log(x_ratio))

    def __getitem__(self, molecule):
        """Overrides getitem so that EquChemTable['molecule'] directly accesses 
        the database for that molecule.
        """
        if molecule not in self.vol_mix_ratio:
            raise KeyError('The requested molecule is not available.')
        return self.vol_mix_ratio[molecule]

    def vmr(self, logP, T, mol):
        """Interpolates a single molecule on a logP-T value
        """
        return np.exp(self.logx_interp[mol](logP,T,grid=False)[0])

    def interpolate_vmr(self, logp_array=None,t_array=None, mols=None,grid=False):
        """Interpolates all molecules in mols on a logP-T grid
        """
        res={}
        for mol in mols:
            res[mol]=np.exp(self.logx_interp[mol](logp_array, t_array, grid=grid))
        return res

#    def vmr_logPT(self, logPgrid, tgrid, mols, is_sorted=False):
#        """Interpolates all molecules in mols on a logP-T grid
#        """
#        res={}
#        #lP,T=np.meshgrid(logPgrid,tgrid)
#        for mol in mols:
#            #res[mol]=10**self.logx_interp[mol](logPgrid.reshape((self.Np,1)),tgrid)
#            res[mol]=np.exp(self.logx_interp[mol](logPgrid, tgrid, \
#                assume_sorted=is_sorted).transpose())
#        return res



class InterpolationChemistry(object):
    """Chemistry model interpolating Volume Mixing Ratios based on a datafile
    provided by V. Parmentier.
    """

    def __init__(self, filename=None, t_min=260):
        """

        Parameters
        ----------
            filename: str
                The path to the datafile.

            t_min: float
                minimal temperature to consider.
                Cannot be below 260K because not enough data in the file.

        """
        self.df = pd.read_csv(filename,sep=r"\s{1,}", engine='python')
        self.df = self.df[self.df["T"] >= t_min]
        self.tgrid = np.unique(self.df["T"])
        self.logpgrid = np.log10(np.exp(np.unique(self.df["P"])))+2.
        #self.logpgrid = np.unique(self.df["P"])
        self.data = {}

        for gas in self.df.columns:
            if gas in ['T','P']: continue
            self.data[gas] = np.empty(self.logpgrid.shape + self.tgrid.shape)
            for i_t in range(self.tgrid.size):
                for i_p in range(self.logpgrid.size):
                    if gas=='e-':
                        pass
                    self.data[gas][i_p, i_t] = self.df[gas].iloc[i_t * len(self.logpgrid) + i_p]

    def compute_vmr(self, molecule = None, logp_array = None, t_array = None):
        """

        Parameters
        ----------
            molecule: str
                Name of molecule to deal with
            logp_array: array
                Array of log10 pressures (Pa)
            t_array: array
                Array of temperatures (K)

        """
        if (molecule is None) or (logp_array is None) or (t_array is None):
            raise RuntimeError('Missing keyword argument.')
        logp_array=np.array(logp_array, dtype=float)
        t_array=np.array(t_array, dtype=float)
        i_p,weight_p = interp_ind_weights(logp_array, self.logpgrid)
        i_t,weight_t = interp_ind_weights(t_array, self.tgrid)

        p1t1=self.data[molecule][i_p,  i_t  ]
        p0t1=self.data[molecule][i_p-1,i_t  ]
        p1t0=self.data[molecule][i_p,  i_t-1]
        p0t0=self.data[molecule][i_p-1,i_t-1]

        return np.exp(bilinear_interpolation_array(p0t0, p1t0, p0t1, p1t1, weight_p, weight_t))

    def compute_vmr_grid(self, molecule = None, logp_array = None, t_array = None):
        if (molecule is None) or (logp_array is None) or (t_array is None):
            raise RuntimeError('Missing keyword argument.')
        i_p,weight_p = interp_ind_weights(logp_array, self.logpgrid)
        i_t,weight_t = interp_ind_weights(t_array, self.tgrid)
        
        # now put everything on 2D grids.
        tmp=np.meshgrid(i_p, i_t, indexing='ij')
        shape=tmp[0].shape
        i_p_grid=tmp[0].ravel()
        i_t_grid=tmp[1].ravel()
        tmp=np.meshgrid(weight_p, weight_t, indexing='ij')
        weight_p_grid=tmp[0].ravel()
        weight_t_grid=tmp[1].ravel()
        
        p1t1=self.data[molecule][i_p_grid,  i_t_grid  ]
        p0t1=self.data[molecule][i_p_grid-1,i_t_grid  ]
        p1t0=self.data[molecule][i_p_grid,  i_t_grid-1]
        p0t0=self.data[molecule][i_p_grid-1,i_t_grid-1]
        
        return np.reshape(np.exp(bilinear_interpolation_array(p0t0, p1t0, p0t1, p1t1,
                weight_p_grid, weight_t_grid)), shape)



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Module containing a class to handle a database of CIA tables to compute opacities with it.
"""
import numpy as np

from .atable import Atable
from .util.spectral_object import Spectral_object

class Adatabase(Spectral_object):
    """Class to group :class:`~exo_k.atable.Atable` objects
    and combine them in radiative transfer
    """

    def __init__(self, *str_filters, filenames=None, remove_zeros=False, **kwargs):
        """Initializes aerosol tables and supporting data from a list of filenames.

        Parameters
        ----------
            filenames: list
                List of names (not full path) of the input cia files. 
                The files must be in the global search path.
                
        A local search path can be specified with 'search_path='

        See the options of :class:`~exo_k.atable.Atable` __init__ method.
        """
        self.atables={}
        self.r_eff_unit=None
        self.wns=None
        self.Nw=None
        if filenames is not None:
            for filename in filenames:
                tmp_atable=Atable(*([filename]+list(str_filters)),
                    remove_zeros=remove_zeros, **kwargs)
                self.add_atables(tmp_atable)
    
    def add_atables(self, *atables):
        """Adds new :class:`~exo_k.atable.Atable` objects to a Adatabase.

        Parameters
        ----------
            atables: :class:`~exo_k.atable.Atable`
                As many Atables as you want.
        """
        for atable in atables:
            if self.r_eff_unit is None:
                self.r_eff_unit=atable.r_eff_unit
            elif self.r_eff_unit!=atable.r_eff_unit:
                print('Adatabase units for r_eff: {k}'.format(k=self.r_eff_unit))
                print('{aer} units for cia coeff: {k}'.format(\
                    aer=atable.aerosol_name, k=atable.r_eff_unit))
                raise RuntimeError("""You naughty:
                all Atables in a database must have the same r_eff units""")
            # should ultimately check for wn_unit as well !

            self.atables[atable.aerosol_name]=atable

    @property
    def names(self):
        """Gives the names of the aerosols in the database
        """
        return list(self.atables.keys())

    def __getitem__(self, name):
        """Overrides getitem so that Adatabase['name'] directly accesses 
        the database for that aerosol name.
        """
        if name not in self.atables:
            raise KeyError('The requested aerosol is not available.')
        return self.atables[name]

    def copy(self):
        """Creates a new instance of :class:`CIAdatabase` object and (deep) copies data into it
        """
        res=Adatabase(None)
        for atab in self.atables.values():
            res.add_atables(atab.copy())
        return res

    def __repr__(self):
        """Method to output
        """
        output='The available aerosols are: \n'
        for aer,atab in self.atables.items():
            output+=aer+'->'+atab.filename+'\n'
        if self.wns is not None:
            output+='All tables share a common spectral grid\n'
        else:
            output+='All tables do NOT have common spectral grid\n'
            output+='You will need to run sample before using the database\n'

        return output

    def sample(self, wngrid, remove_zeros=True, use_grid_filter=True,
            sample_all_vars=True):
        """Samples all the Atables in the database on the same wavenumber grid
        to be able to use them in radiative transfer modules.

        .. important::
            For this method, the default options are remove_zeros=True and use_grid_filter=True.
            This ensures that you do not count several contributinns twice when
            you use two different aerosol tables to represent
            the same aerosol type over two different
            spectral window (typically VI and IR)

        Parameters
        ----------
            wngrid : array
                new wavenumber grid (cm-1)

        .. seealso::
            See :func:`exo_k.atable.Atable.sample for further details on options.
        """
        for atable in self.atables.values():
            atable.sample(wngrid, remove_zeros=remove_zeros, use_grid_filter=use_grid_filter,
                sample_all_vars=sample_all_vars)
        self.wns=np.array(wngrid)
        self.Nw=self.wns.size

    def convert_to_mks(self):
        """Converts units of all Cia_tables to MKS.
        """
        first=True
        for atable in self.atables.values():
            atable.convert_to_mks()
            if first:
                self.r_eff_unit=atable.r_eff_unit
        

    def absorption_coefficient(self, aer_reffs_densities,
            wngrid_limit=None, log_interp=None):
        """Computes the absorption coefficient in m^-1 for the whole mix specified 
        (assumes data in MKS).

        Parameters
        ----------
            aer_reffs_densities: dict
                A dictionary with aerosol names as keys and lists containing 2
                floats (or arrays) as values. The values are the particle effective radii
                and number densities.

            wngrid_limit: array, optional
                Smaller and bigger wavenumbers inside which to perform the calculation.

        Returns:
            array:
                The effective cross section coefficient profile for the aerosols (in m^2).
        """
        first=True
        if self.wns is None: raise RuntimeError("""Atables must be sampled
            on the same grid before calling cross_section.
            Please use the sample() method.""")
        for aer, values in aer_reffs_densities.items():
            if aer in self.atables.keys():
                if first:
                    res=self.atables[aer].absorption_coefficient(values[0], values[1],
                        wngrid_limit=wngrid_limit, log_interp=log_interp)
                    first=False
                else:
                    res+=self.atables[aer].absorption_coefficient(values[0], values[1],
                        wngrid_limit=wngrid_limit, log_interp=log_interp)
        if first: # means that no molecule was in the database, we need to initialize res
            res=np.zeros((1, self.wns.size))
        return res

    def optical_properties(self, aer_reffs_densities, wngrid_limit=None,
            log_interp=None, compute_all_opt_prop=True):
        """Computes the optical properties for the mix of aerosols
        (assumes data in MKS).

        Parameters
        ----------
            aer_reffs_densities: dict
                A dictionary with aerosol names as keys and lists containing 2
                floats (or arrays) as values. The values are the particle effective radii
                and number densities.

            wngrid_limit: array, optional
                Smaller and bigger wavenumbers inside which to perform the calculation.

        Returns:
            array:
                The effective cross section coefficient profile for the aerosols (in m^2).
        """
        first=True
        if self.wns is None: raise RuntimeError("""Atables must be sampled
            on the same grid before calling optical_properties.
            Please use the sample() method.""")
        for aer, values in aer_reffs_densities.items():
            if aer in self.atables.keys():
                [k, kscat, g]=self.atables[aer].optical_properties(values[0], values[1],
                        wngrid_limit=wngrid_limit, log_interp=log_interp,
                        compute_all_opt_prop=compute_all_opt_prop)
                k_scat_g = kscat * g
                if first:
                    k_tot = k
                    kscat_tot = kscat
                    g_tot = k_scat_g
                    first=False
                else:
                    k_tot += k 
                    kscat_tot += kscat
                    g_tot += k_scat_g
        if first: # means that no molecule was in the database, we need to initialize res
            res=[np.zeros((1,self.wns.size)), np.zeros((1,self.wns.size)), np.zeros((1,self.wns.size))]
        else:
            if compute_all_opt_prop: g_tot /= kscat_tot
            res = [k_tot, kscat_tot, g_tot]
        return res






#!/usr/bin/env python
"""
@author: Aurelien Falco
"""
from distutils.core import setup
from setuptools import find_packages

packages = find_packages(exclude=('tests', 'doc'))
provides = ['exo_k', ]

requires = []

python_requires='>=3'

install_requires = ['numpy',
                    'scipy',
                    'numba>=0.49',
                    'astropy',
                    'h5py',
                    'matplotlib',
                    'panda'
                    ]

console_scripts = ['']

entry_points = {'console_scripts': console_scripts, }

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Console',
    'Framework :: Jupyter',
    'Intended Audience :: Developers',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Operating System :: MacOS',
    'Operating System :: POSIX',
    'Operating System :: POSIX :: Linux',
    'Operating System :: Unix',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Astronomy',
    'Topic :: Scientific/Engineering :: Atmospheric Science',
    'Topic :: Scientific/Engineering :: Physics',
    'Topic :: Software Development :: Libraries',
]

# Handle versioning
version = '1.1.0'

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(name='exo_k',
      author='Jeremy Leconte',
      author_email='jeremy.leconte@u-bordeaux.fr',
      license="GPLv3",
      version=version,
      description='Library to handle radiative opacities from various sources for atmospheric applications',
      classifiers=classifiers,
      packages=packages,
      long_description=long_description,
      url='https://forge.oasu.u-bordeaux.fr/jleconte/exo_k-public',
      long_description_content_type="text/markdown",
      keywords = ['opacities','cross sections','correlated-k',
        'spectra','atmosphere','atmospheric',
        'exopanet', 'radiative transfer'],
      include_package_data=True,
      entry_points=entry_points,
      provides=provides,
      requires=requires,
      install_requires=install_requires,
      extras_require={},
      )



"""
@author: jeremy leconte

A class with some basic functions for all objects with a spectral dimension
"""
import numpy as np

class Spectral_object(object):
    """A class with some basic functions for all objects with a spectral dimension
    """

    @property
    def wls(self):
        """Returns the wavelength array for the bin centers
        """
        if self.wns is not None:
            return 10000./self.wns
        else:
            return None

    @property
    def wledges(self):
        """Returns the wavelength array for the bin edges
        """
        if self.wnedges is not None:
            return 10000./self.wnedges
        else:
            return None

    @property
    def wnrange(self):
        """Returns the limits of the wavenumber range.

        First tries with wnedges (for Ktables) and then wns (Xtables).
        """
        if self.wnedges is not None:
            return self.wnedges[[0,-1]]
        elif self.wns is not None:
            return self.wns[[0,-1]]
        else:
            return None

    @property
    def wlrange(self):
        """Returns the limits of the wavelength range
        """
        if (self.wnedges is None) and (self.wns is None):
            return None
        else:
            return np.sort(10000./self.wnrange)



# -*- coding: utf-8 -*-
"""
Created in Jan 2021

@author: jeremy leconte
"""
import numpy as np
import numba

jit=False

@numba.jit(nopython=True, fastmath=True, cache=True)
def solve_2stream_nu_xsec(source_nu, dtau_nu, omega0_nu, g_asym_nu,
                flux_top_dw_nu, mu0=2./3., alb_surf=0., verbose=False, flux_at_level=False):
    """Deals with the spectral axis
    """
    NLEV, NW = source_nu.shape
    flux_up=np.zeros((NLEV, NW))
    flux_dw=np.zeros((NLEV, NW))
    flux_net=np.zeros((NLEV, NW))
    kernel=np.zeros((NLEV, NLEV, NW))
    for iW in range(NW):
        flux_up[:,iW], flux_dw[:,iW], flux_net[:,iW], kernel[:,:,iW] = \
            solve_2stream(source_nu[:,iW], dtau_nu[:,iW],
                omega0_nu[:,iW], g_asym_nu[:,iW],
                mu0=mu0, flux_top_dw=flux_top_dw_nu[iW],
                alb_surf=alb_surf, verbose=verbose)
    return flux_up, flux_dw, flux_up-flux_dw, kernel
    
@numba.jit(nopython=True, fastmath=True, cache=True)
def solve_2stream_nu_corrk(source_nu, dtau_nu, omega0_nu, g_asym_nu,
                flux_top_dw_nu, mu0=2./3., alb_surf=0., verbose=False, flux_at_level=False):
    """Deals with the spectral axis
    """
    NLEV, NW = source_nu.shape
    NG = dtau_nu.shape[-1]
    flux_up=np.zeros((NLEV, NW, NG))
    flux_dw=np.zeros((NLEV, NW, NG))
    flux_net=np.zeros((NLEV, NW, NG))
    kernel=np.zeros((NLEV, NLEV, NW, NG))
    for iW in range(NW):
        for iG in range(NG):
            flux_up[:,iW, iG], flux_dw[:,iW,iG], flux_net[:,iW,iG], kernel[:,:,iW,iG] = \
                solve_2stream(source_nu[:,iW], dtau_nu[:,iW,iG], 
                    omega0_nu[:,iW,iG], g_asym_nu[:,iW,iG],
                    mu0=mu0, flux_top_dw=flux_top_dw_nu[iW],
                    alb_surf=alb_surf, verbose=verbose)
    return flux_up, flux_dw, flux_up-flux_dw, kernel


@numba.jit(nopython=True, fastmath=True, cache=True)
def solve_2stream(source, dtau, omega0, g_asym, mu0=2./3., flux_top_dw=0.,
            alb_surf=0., verbose=False):
    """
    Inherited from ExoRem (Blain et al. 2020):
    https://gitlab.obspm.fr/dblain/exorem/-/blob/master/src/fortran/exorem/radiative_transfer.f90

    PURPOSE:                                                        

    THIS subroutine COMPUTES THE UPWARD, DOWNWARD AND NET THERMAL   
    FLUX IN AN INHOMOGENEOUS ABSORBING, SCATTERING ATMOSPHERE.      
    THE TWO_STREAM APPROXIMATION (QUADRATURE WITH COS OF            
    AVERAGE ANGLE = mu0) IS USED TO FIND THE DIFFUSE   	        
    REFLECTIVITY AND TRANSMISSIVITY AND THE TOTAL UPWARD AND        
    DOWNWARD FLUXES FOR EACH OF THE NLAY HOMOGENEOUS LAYERS.        
    THE ADDING METHOD IS then USED TO COMBINE THESE LAYERS.  IF     
    ANY LAYER IS THICKER THAN DTAU = *EMAX1*, IT IS ASSUMED TO BE   
    SEMI-INFINITE.  LAYERS THICKER THAN DTAU = *EMAX2*, ARE TREATED 
    AS INFINITE LAYERS.                                             

    NOTE: TO ACCOUNT FOR A DIFFUSE FLUX AT THE TOP OF THE ATMOS-    
          PHERE, THE USER MUST SET flux_dw(1) EQUAL TO THAT VALUE.     
          FOR NO DOWNWARD DIFFUSE FLUX AT THE TOP OF THE ATMOSPHERE,
          THE USER MUST INITIALIZE flux_dw(1) TO ZERO IN THE CALLING   
          PROGRAM.                                                  


    Parameters
    ----------
        source: array
            PLANCK function AT EACH LEVEL FROM TOP OF THE          
            ATMOSPHERE (L=0) TO THE SURFACE (L=NLAY) (NLAY+1 VALUES)  
            JL: actually seems to be pi times the planck function
            as defined in the rest of the code. 
        dtau: array
            ARRAY OF NORMAL INCIDENCE OPTICAL DEPTHS IN EACH       
            HOMOGENEOUS MODEL LAYER. (NLAY VALUES)                 
        omega0: array
            ARRAY OF SINGLE SCATTERING ALBEDOS FOR EACH HOMO-      
            GENEOUS MODEL LAYER. (NLAY VALUES)                     
        g_asym: array
            ARRAY OF ASSYMETRY parameterS FOR EACH HOMOGENEOUS     
            MODEL LAYER. (NLAY VALUES)                             
    
    Returns
    -------                                                        

        flux_up: array
            UPWARD FLUX AT NLAY+1 LAYER BOUNDARIES.                 
            (flux_up(L) REFERS TO THE UPWARD FLUX AT THE TOP          
            OF LAYER L)                                           
        flux_dw: array
            DOWNWARD FLUX AT NLAY+1 LAYER BOUNDARIES.               
            (flux_dw(L) REFERS TO THE DOWNWARD FLUX AT THE BOTTOM     
            OF LAYER L-1) 
        flux_up-flux_dw: array
            Net flux at the same levels

    Other Parameters
    ----------------
        mu0: float
            Cos of quadrature angle (Original mu0 was 2/3.)
        flux_top_dw: float
            Diffuse downward flux at upper interface.
    """
#            se atmosphere, only : n_levels
#
#            implicit none
#
#            
#            ********** COMMON BLOCKS USED IN DELTA-EDDINGTON ROUTINES.
#            
#            integer, intent(in) :: nlay
#            doubleprecision, intent(in) :: dtau(n_levels - 1),
#                   g_asym(n_levels - 1), source(n_levels)
#
#            doubleprecision, intent(inout) :: omega0(n_levels - 1)
#
#            doubleprecision, intent(out) :: flux_up(n_levels), flux_dw(n_levels),
#                 DKERNEL(n_levels, n_levels)
#
#            doubleprecision :: source2(n_levels), OMP(n_levels - 1), dtauP(n_levels - 1), 
#                RU(n_levels), &
#                DD(n_levels), DFLUX(n_levels - 1), G1(n_levels - 1), G2(n_levels - 1), &
#                DU(n_levels - 1), UFL(n_levels), DFL(n_levels), UFLUX(n_levels), 
#                EKT(n_levels - 1), &
#                SK(n_levels - 1), RL(n_levels), temperatures_layers(n_levels), RS(n_levels)
#
#            integer :: j, l, np2, nlev
#            doubleprecision :: skt, prec, emax1, emax2, alb, mu0, denom, e2ktm, emis,
#                flux_top_dw
    PREC = 1.e-10
    EMAX1 = 8.
    EMAX2 = 24.
    #EMAX1, EMAX2: optical depth at which layers are treated as semi infinite
    #        and infinite (resp). 
    NLAY = dtau.size
    NLEV = NLAY + 1
    NP2 = NLAY + 2

    dtauP=np.zeros_like(dtau)
    OMP=np.zeros_like(dtau)
    G1=np.zeros_like(dtau)
    G2=np.zeros_like(dtau)
    SK=np.zeros_like(dtau)
    DU=np.zeros_like(dtau)
    EKT=np.zeros_like(dtau)
    DFLUX=np.zeros_like(dtau)

    RL=np.zeros_like(source)
    RS=np.zeros_like(source)
    RU=np.zeros_like(source)
    DD=np.zeros_like(source)
    DFL=np.zeros_like(source)
    UFLUX=np.zeros_like(source)
    UFL=np.zeros_like(source)
    source2=np.zeros_like(source)
    temperatures_layers=np.zeros_like(source)
    flux_up=np.zeros_like(source)
    flux_dw=np.zeros_like(source)

    DKERNEL=np.zeros((NLEV,NLEV))
    #
    #****   SCALE THE OPTICAL DEPTHS, SINGLE SCATTERING ALBEDOS AND THE
    #       SCATTERING ASSYMETRY FACTORS FOR USE IN THE TWO-STREAM
    #       APPROXIMATION.  INITIALIZE OTHER QUANTITIES.  USE WISCOMBE'S
    #       TRICK TO SUBTRACT A SMALL VALUE FROM THE SINGLE SCATTERING
    #       FOR THE CASE OF A CONSERVATIVE ATMOSPHERE
    #
    for L in range(NLAY):
        if (1. - omega0[L])  <  PREC: omega0[L] = 1. - PREC
        dtauP[L] = (1. - omega0[L] * g_asym[L]) * dtau[L]
        OMP[L] = (1. - g_asym[L]) * omega0[L] / (1. - omega0[L] * g_asym[L])
        G1[L] = (1. - 0.5 * OMP[L]) / mu0
        G2[L] = 0.5 * OMP[L] / mu0
        SK[L] = np.sqrt(G1[L] * G1[L] - G2[L] * G2[L])
    #
    #**** DIFFUSE TRANSMITTANCE AND REFLECTANCE OF EACH HOMOGENEOUS LAYER
    #
    #     THE EQUATIONS FOR RL AND temperatures_layers WERE DERIVED BY SOLVING THE HOMOGENEOUS
    #     PART OF THE EQUATION OF TRANSFER (IE. NO SOURCE TERM)
    #
    for L in range(NLAY):
        SKT = SK[L] * dtauP[L]
        #print(L)
        if SKT  > EMAX2 :
        #
        #     INFINITE LAYERS
        #
            #print('EMAX2', L, SKT, G1[L], SK[L], (G1[L] + SK[L]))
            RL[L] = G2[L] / (G1[L] + SK[L])
            temperatures_layers[L] = 0.e0
            continue
        EKT[L] = np.exp(SKT)
        if SKT  >  EMAX1 :
        #
        #     SEMI-INFINITE LAYERS
        #
            #print('EMAX1', L, SKT, (G1[L] + SK[L]),(EKT[L] * (G1[L] + SK[L])))
            RL[L] = G2[L] / (G1[L] + SK[L])
            temperatures_layers[L] = 2.e0 * SK[L] / (EKT[L] * (G1[L] + SK[L]))
            continue
        E2KTM = EKT[L] * EKT[L] - 1.
        DENOM = G1[L] * E2KTM + SK[L] * (E2KTM + 2.e0)

        RL[L] = G2[L] * E2KTM / DENOM
        temperatures_layers[L] = 2.e0 * SK[L] * EKT[L] / DENOM
    #
    #****   SET THE "REFLECTIVITY", "TRANSMISSIVITY" AND "EMISSIVITY" OF
    #       THE "SURFACE" ASSUMING SEMI-INFINITE LAYER WITH SAME OMP
    #       AS LAYER NLAY. FOR "EMISSIVITY", ASSUME SAME dB/dTau AS LAYER NLAY
    #ALB = G2[NLAY-1] / (G1[NLAY-1] + SK[NLAY-1])
    #RL[NLEV-1] = (1. - ALB)
    #EMIS = 1. - ALB + (1. + ALB) * mu0 * \
    #    (1. - source[NLEV - 2] / source[NLEV-1]) / dtauP[NLAY-1]
    ##JL21 below, I put a new boundary condition at the bottom to mimick a dark surface.
    ALB=alb_surf
    RL[NLEV-1]=1.
    EMIS=1.-ALB
    temperatures_layers[NLEV-1] = 0.e0
    #
    #****   USE ADDING METHOD TO FIND THE REFLECTANCE AND TRANSMITTANCE
    #       OF COMBINED LAYERS.  ADD DOWNWARD FROM THE TOP AND UPWARD
    #       FROM THE BOTTOM AT THE SAME TIME.
    #
    RS[0] = RL[0]
    RU[NLEV-1] = ALB

    for L in range(NLAY):
        DD[L] = 1. / (1. - RS[L] * RL[L + 1])
        RS[L + 1] = RL[L + 1] + temperatures_layers[L + 1] * \
            temperatures_layers[L + 1] * RS[L] * DD[L]
        DU[NLEV - L - 2] = 1. / (1. - RL[NLEV - L - 2] * RU[NP2 - L - 2])
        RU[NLEV - L - 2] = RL[NLEV - L - 2] + temperatures_layers[NLEV - L - 2] * \
            temperatures_layers[NLEV - L - 2] * RU[NP2 - L - 2] * DU[NLEV - L - 2]
    #
    #****   COMPUTE THE UPWARD AND DOWNWARD FLUX FOR EACH HOMOGENEOUS LAYER
    #
    #**** LOOP OVER J FOR KERNEL
    for J in range(NLEV + 1):
        DFL[NLEV-1] = 0.e0
        if J <= NLEV-1:
            flux_dw[0] = 0.e0
            for L in range(NLEV):
                DKERNEL[J, L] = 0.e0
                source2[L] = 0.e0
            source2[J] = 1.
        else:
            flux_dw[0] = flux_top_dw
            for L in range(NLEV):
                source2[L] = source[L]
        UFL[NLEV-1] = EMIS * source2[NLEV-1]
        for L in range(NLAY):
            UFL[L], DFL[L]=_DEDIR1(source2[L], source2[L + 1], omega0[L],
                                g_asym[L], dtau[L], mu0)
        #
        #****   USE ADDING METHOD TO FIND UPWARD AND DOWNWARD FLUXES
        #       FOR COMBINED LAYERS.  START AT TOP
        #
        DFLUX[0] = DFL[0] + temperatures_layers[0] * flux_dw[0]

        for L in range(NLAY-1):
            DFLUX[L + 1] = temperatures_layers[L + 1] * \
                (RS[L] * (RL[L + 1] * DFLUX[L] + UFL[L + 1]) * DD[L] + DFLUX[L]) + DFL[L + 1]
        flux_dw[NLEV-1] = (DFLUX[NLAY-1] + RS[NLAY-1] * EMIS * source2[NLEV-1]) / \
            (1. - RS[NLAY-1] * ALB)
        #
        #****   USE ADDING METHOD TO FIND UPWARD AND DOWNWARD FLUXES
        #       FOR COMBINED LAYERS.  START AT BOTTOM.
        #
        UFLUX[NLEV-1] = UFL[NLEV-1]
        for L in range(NLAY):
            UFLUX[NLEV - L - 2] = temperatures_layers[NLEV - L - 2] * (UFLUX[NP2 - L - 2] + \
                    RU[NP2 - L - 2] * DFL[NLEV - L - 2]) * DU[NLEV - L - 2] + \
                    UFL[NLEV - L - 2]
        #
        #****   FIND THE TOTAL UPWARD AND DOWNWARD FLUXES AT INTERFACES
        #       BETWEEN INHOMOGENEOUS LAYERS.
        #
        flux_up[0] = UFLUX[0] + RU[0] * flux_dw[0]
        flux_up[NLEV-1] = ALB * flux_dw[NLEV-1] + EMIS * source2[NLEV-1]
        for L in range(NLAY-1):
            flux_up[NLEV - L - 2] = (UFLUX[NLEV - L - 2] + RU[NLEV - L - 2] * \
                    DFLUX[NLAY - L - 2]) / (1. - RU[NLEV - L - 2] * \
                    RS[NLAY - L - 2])
        for L in range(NLAY-1):
            flux_dw[L + 1] = DFLUX[L] + RS[L] * flux_up[L + 1]
        #
        if J <= NLEV-1:
            for L in range(NLEV):
                DKERNEL[J, L] = flux_up[L] - flux_dw[L]

    return flux_up, flux_dw, flux_up-flux_dw, DKERNEL

@numba.jit(nopython=True, fastmath=True, cache=True)
def _DEDIR1(BATM1, BATM2, AIR, GIR, TAU, mu0):
    """            
            CCCCCCCCCCCCCCCCCCCCCCCCCCC  D E D I R 1  CCCCCCCCCCCCCCCCCCCCCCCCCCCCC
            C                                                                    CC
            C    PURPOSE :                                                       CC
            C                                                                    CC
            C    THIS subroutine USES THE TWO-STREAM ROUTINE (QUADRATURE WITH    CC
            C    COS OF AVERAGE ANGLE = mu0) TO FIND THE UPWARD AND DOWNWARD    CC
            C    THERMAL FLUXES EMITTED FROM A HOMOGENEOUS LAYER.                CC
            C                                                                    CC
            C    AUTHORS:  DAVID PAIGE AND DAVID CRISP                           CC
            C                                                                    CC
            C    INPUT:                                                     CC
            C                                                                    CC
            C    BTATM1 IS ATMOSPHERIC EMISSION at level L (ARBITRARY UNITS)     CC
            C    BTATM2 IS ATMOSPHERIC EMISSION at level L+1 (ARBITRARY UNITS)   CC
            C    AIR IS IR SINGLE SCATTERING ALBEDO                              CC
            C    GIR IS IR ASYMMETRY parameter                                   CC
            C    TAU IS OPTICAL DEPTH (EXTINCTION = ABSORPTION + SCATTERING)     CC
            C                                                                    CC
            C    OUTPUT :                                                   CC
            C                                                                    CC
            C    FUPTOP IS UPWARD FLUX AT TOP OF ATM (SAME UNITS AS BTATM1 or 2) CC
            C    FDNBOT IS DOWNWARD FLUX AT SURFACE                              CC
            C    FUPBOT IS UPWARD FLUX AT SURFACE                                CC
            C                                                                    CC
            CCCCCCCCCCCCCCCCCCCCCCCCCCC  D E D I R 1  CCCCCCCCCCCCCCCCCCCCCCCCCCCCC
    """        
#            doubleprecision, intent(in) :: BATM1, BATM2, air, gir, tau
#            doubleprecision, intent(out) :: fuptop, fdnbot
#
#            doubleprecision :: mu0, c1, c2, cap, dair, db, dtauIR, emkt, \
#                          epkt, omttp, opttp, v1, v2, v3, v4, v5, v6
    DAIR = AIR * (1. - GIR) / (1. - GIR * AIR)
    CAP = np.sqrt(1. - DAIR) / mu0
    OPTTP = np.sqrt(1. - 0.5 * DAIR + np.sqrt(1. - DAIR))
    OMTTP = np.sqrt(1. - 0.5 * DAIR - np.sqrt(1. - DAIR))
    dtauIR = (1. - GIR * AIR) * TAU
    db = (BATM2 - BATM1) * mu0 / dtauIR

    if CAP * dtauIR < 24.:
        #
        #****     THE LAYER THICKNESS IS FINITE
        #
        EPKT = np.exp(+CAP * dtauIR)
        EMKT = 1. / EPKT
        #
        #****     SET TOP B.C.
        #
        V1 = OPTTP
        V2 = OMTTP
        V3 = -(BATM1 - db)
        #
        #****     SET LOWER B.C.
        #
        V4 = EMKT * OMTTP
        V5 = EPKT * OPTTP
        V6 = -(BATM2 + db)
        #
        #****     SOLVE SYSTEM OF EQUATIONS
        #
        C1 = (V3 * V5 - V6 * V2) / (V1 * V5 - V4 * V2)
        C2 = (V1 * V6 - V3 * V4) / (V1 * V5 - V4 * V2)

        FUPTOP = C1 * OMTTP + C2 * OPTTP + BATM1 + db
        FDNBOT = C1 * EMKT * OPTTP + C2 * EPKT * OMTTP + BATM2 - db

    else:
        #
        #****     ASSUME LAYER IS SEMI-INFINITE.
        #
        FUPTOP = BATM1 * (1. - OMTTP / OPTTP) + db * (1. + OMTTP / OPTTP)
        FDNBOT = BATM2 * (1. - OMTTP / OPTTP) - db * (1. + OMTTP / OPTTP)
    return FUPTOP, FDNBOT



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
from .util.spectral_object import Spectral_object

class Aerosols(Spectral_object):
    """Dict-like class to handle gas composition (with background gas) and molar mass.

    If `logp_array`, `t_array`, and radiative databases are provided, :any:`cross_section`
    can be used to compute the opacity of the gas
    """

    def __init__(self, aer_reffs_densities=None, a_database=None):
        """__init__ Instantiates
        an aerosols object.
        """
        self.set_a_database(a_database=a_database)
        self.set_aer_reffs_densities(aer_reffs_densities=aer_reffs_densities)
        self._wn_range=None

    def set_aer_reffs_densities(self, aer_reffs_densities=None):
        if aer_reffs_densities is not None:
            self.aer_reffs_densities = aer_reffs_densities

    def set_a_database(self, a_database=None):
        """Change the radiative database attached to the current instance of aerosols

        Parameters
        ----------
            a_database: :class:`~exo_k.kdatabase.Kdatabase` object
                New Adatabase to use.
        """
        self.adatabase=a_database
        if self.adatabase is not None and self.adatabase.r_eff_unit != 'm' :
            print("""
            You're being Bad!!! You are trying *NOT* to use MKS units
            for the effective radii in the aerosol database!!!""")
            raise RuntimeError("Bad units in the Adatabase used with aerosols.")

    def _compute_spectral_range(self, wn_range=None, wl_range=None):
        """Converts an unordered spectral range in either wavenumber or wavelength
        in an ordered wavenumber range.

        Parameters
        ----------
            wn_range: list or array of size 2
                Minimum and maximum wavenumber (in cm^-1).
            wl_range: list or array of size 2
                Minimum and maximum wavelength (in micron)
        """
        if wl_range is not None:
            if wn_range is not None:
                print('Cannot specify both wl and wn range!')
                raise RuntimeError()
            else:
                _wn_range=np.sort(10000./np.array(wl_range))
        else:
            if wn_range is not None:
                _wn_range=np.sort(np.array(wn_range))
            else:
                _wn_range=self._wn_range
        return _wn_range

    def optical_properties(self, aer_reffs_densities=None, wl_range=None, wn_range=None,
            log_interp=True, compute_all_opt_prop=True, **kwargs):
        """bla
        """
        if self.adatabase is None: raise RuntimeError("""
            a_database not provided. 
            Use the a_database keyword during initialization or use the set_a_database method.""")
        if self.adatabase.wns is None: raise RuntimeError("""
            All tables in the Adatabase should have the same wavenumber grid to proceed.
            You should probably use sample().""")
        if aer_reffs_densities is not None:
            self.set_aer_reffs_densities(aer_reffs_densities=aer_reffs_densities)

        local_wn_range=self._compute_spectral_range(wl_range=wl_range, wn_range=wn_range)

        [k, k_scat, g] = self.adatabase.optical_properties(self.aer_reffs_densities,
            wngrid_limit=local_wn_range, log_interp=log_interp,
            compute_all_opt_prop=compute_all_opt_prop)
        return [k, k_scat, g]
    
    def absorption_coefficient(self, aer_reffs_densities=None, wl_range=None, wn_range=None,
            log_interp=True, **kwargs):
        """bla
        """
        if self.adatabase is None: raise RuntimeError("""
            a_database not provided. 
            Use the a_database keyword during initialization or use the set_a_database method.""")
        if self.adatabase.wns is None: raise RuntimeError("""
            All tables in the Adatabase should have the same wavenumber grid to proceed.
            You should probably use sample().""")
        if aer_reffs_densities is not None:
            self.set_aer_reffs_densities(aer_reffs_densities=aer_reffs_densities)

        local_wn_range=self._compute_spectral_range(wl_range=wl_range, wn_range=wn_range)

        return self.adatabase.absorption_coefficient(self.aer_reffs_densities,
            wngrid_limit=local_wn_range, log_interp=log_interp)




# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
__init__ module to load util subpackage
"""
from .radiation import *
from .cst import *
from .spectrum import Spectrum
from .filenames import *
from .molar_mass import Molar_mass
from .user_func import *
